import collections
import copy
import datetime
import functools
import re

import bson
import sortedcontainers

from .common import support_alert, ASCENDING, DESCENDING, MetaStorageObject
from .cursor import Cursor, _validate_sort
from .errors import (MongitaError, MongitaNotImplementedError, DuplicateKeyError,
                     InvalidName, OperationFailure)
from .read_concern import ReadConcern
from .results import InsertOneResult, InsertManyResult, DeleteResult, UpdateResult
from .write_concern import WriteConcern

_SUPPORTED_FILTER_OPERATORS = ('$in', '$eq', '$gt', '$gte', '$lt', '$lte', '$ne', '$nin')
_SUPPORTED_UPDATE_OPERATORS = ('$set', '$inc', '$push')
_DEFAULT_METADATA = {
    'options': {},
    'indexes': {},
    '_id': str(bson.ObjectId()),
}

# FROM docs.mongodb.com/manual/reference/bson-type-comparison-order/#comparison-sort-order
SORT_ORDER = {
    int: b'\x02',
    float: b'\x02',
    str: b'\x03',
    object: b'\x04',
    list: b'\x05',
    bytes: b'\x06',
    bson.ObjectId: b'\x07',
    bool: b'\x08',
    datetime.datetime: b'\t',
    re.Pattern: b'\n',
}


def _validate_filter(filter):
    """
    Validate the 'filter' parameter.
    This is near the top of most public methods.

    :param filter dict:
    :rtype: None
    """
    if not isinstance(filter, dict):
        raise MongitaError("The filter parameter must be a dict, not %r" % type(filter))
    for k in filter.keys():
        if not isinstance(k, str):
            raise MongitaError("Filter keys must be strings, not %r" % type(filter))
    _id = filter.get('_id')
    if _id:
        if not isinstance(_id, (bson.ObjectId, str, dict)):
            raise MongitaError("If present, the '_id' filter must be a bson ObjectId, string, or a dict")
    for query_ops in filter.values():
        if isinstance(query_ops, dict):
            for op in query_ops.keys():
                if op.startswith('$') and op not in _SUPPORTED_FILTER_OPERATORS:
                    raise MongitaError(
                        "Mongita does not support %r. These filter operators are "
                        "supported: %r" % (op, _SUPPORTED_FILTER_OPERATORS))


def _validate_update(update):
    """
    Validate the 'update' parameter.
    This is near the top of the public update methods.

    :param update dict:
    :rtype: None
    """
    if not isinstance(update, dict):
        raise MongitaError("The update parameter must be a dict, not %r" % type(update))
    for k in update.keys():
        if k in _SUPPORTED_UPDATE_OPERATORS:
            continue
        if k.startswith('$'):
            raise MongitaNotImplementedError(
                "Mongita does not support %r. These update operators are "
                "supported: %r." % (k, _SUPPORTED_UPDATE_OPERATORS))
        raise MongitaError(
            "In update operations, you must use one of the supported "
            "update operators %r." % (_SUPPORTED_UPDATE_OPERATORS,))
    for update_dict in update.values():
        if not isinstance(update_dict, dict):
            raise MongitaError("If present, the update operator must be a dict, "
                               "not %r" % type(update_dict))
        _id = update_dict.get('_id')
        if _id:
            if not isinstance(_id, (str, bson.ObjectId)):
                raise MongitaError("The update _id must be a bson ObjectId or a string")


def _validate_doc(doc):
    """
    Validate the 'doc' parameter.
    This is near the top of the public insert / replace methods.

    :param doc dict:
    :rtype: None
    """

    if not isinstance(doc, dict):
        raise MongitaError("The document must be a dict, not %r" % type(doc))
    _id = doc.get('_id')
    if _id:
        if not isinstance(_id, (bson.ObjectId, str)):
            raise MongitaError("The document _id must be a bson ObjectId, a string, or not present")
    for k in doc.keys():
        if not k or k.startswith('$'):
            raise InvalidName("All document keys must be truthy and cannot start with '$'.")


def _overlap(iter_a, iter_b):
    """
    Return if there is any overlap between iter_a and iter_b
    from https://stackoverflow.com/questions/3170055

    :param iter_a list:
    :param iter_b list:
    :rtype: bool
    """
    return not set(iter_a).isdisjoint(iter_b)


def _doc_matches_agg(doc_v, query_ops):  # noqa: C901
    """
    Return whether an individual document value matches a dict of
    query operations. Usually there will be one query_op but sometimes there
    are many.

    e.g. collection.find({'path.to.doc_v': {'$query_op': query_val}})

    The loop returns False whenever we know for sure that the document is
    not part of the query. At the end return True

    :param doc_v: The value in the doc to compare against
    :param query_ops {$query_op: query_val}:
    :returns: Whether the document value matches all query operators
    :rtype: bool
    """
    if any(k.startswith('$') for k in query_ops.keys()):
        for query_op, query_val in query_ops.items():
            if query_op == '$eq':
                if doc_v != query_val:
                    return False
            elif query_op == '$ne':
                if doc_v == query_val:
                    return False
            elif query_op == '$in':
                if not isinstance(query_val, (list, tuple, set)):
                    raise MongitaError("'$in' requires an iterable")
                if not ((isinstance(doc_v, list) and _overlap(doc_v, query_val))
                        or (doc_v in query_val)):
                    return False
            elif query_op == '$nin':
                if not isinstance(query_val, (list, tuple, set)):
                    raise MongitaError("'$nin' requires an iterable")
                if (isinstance(doc_v, list) and _overlap(doc_v, query_val)) \
                        or (doc_v in query_val):
                    return False
            elif query_op == '$lt':
                try:
                    if doc_v >= query_val:
                        return False
                except TypeError:
                    return False
            elif query_op == '$lte':
                try:
                    if doc_v > query_val:
                        return False
                except TypeError:
                    return False
            elif query_op == '$gt':
                try:
                    if doc_v <= query_val:
                        return False
                except TypeError:
                    return False
            elif query_op == '$gte':
                try:
                    if doc_v < query_val:
                        return False
                except TypeError:
                    return False
            # agg_k check is in _validate_filter
        return True
    else:
        return doc_v == query_ops


def _doc_matches_slow_filters(doc, slow_filters):
    """
    Given an entire doc, return whether that doc matches every filter item in the
    slow_filters dict. A slow_filter is just the set of filters that we didn't
    have an index for.

    :param doc dict:
    :param slow_filters dict:
    :rtype: bool
    """
    for doc_key, query_ops in slow_filters.items():
        if isinstance(query_ops, dict):
            doc_v = _get_item_from_doc(doc, doc_key)
            if _doc_matches_agg(doc_v, query_ops):
                continue
            return False

        item_from_doc = _get_item_from_doc(doc, doc_key)
        if isinstance(item_from_doc, list) and query_ops in item_from_doc:
            continue
        if item_from_doc == query_ops:
            continue
        return False
    return True


def _ids_given_irange_filters(matched_keys, idx, **kwargs):
    """
    Given an existing set of matched_keys, a SortedDict (idx), and
    a set of kwargs to apply to SortedDict.irange,
    return all keys that match both the irange and the existing matched_keys

    :param matched_keys set:
    :param idx sortedcontainers.SortedDict:
    :param kwargs dict: irange filters
    :rtype set:
    """
    clean_idx_key = kwargs.get('minimum') or kwargs.get('maximum')
    # if 'minimum' in kwargs:
    #     kwargs['maximum'] = (bytes([ord(kwargs['minimum'][0]) + 1]), None)
    # if 'maximum' in kwargs:
    #     kwargs['minimum'] = (bytes([ord(kwargs['maximum'][0]) - 1]), None)
    ret = set(idx.irange(**kwargs))
    ret = set(key for key in ret if key[0] == clean_idx_key[0])

    return set.intersection(matched_keys, ret)


def _idx_filter_sort(query_op_tup):
    """
    For performance, the order of filtering matters. It's best to do
    equality first before comparsion. Not-equal should be last becase a lot
    of values are liable to be returned.
    In and nin vary in how much they matter so go with not-equal

    :param query_op_tup (query_op, query_val):
    :rtype: bool
    """
    query_op, _ = query_op_tup
    return (query_op == '$eq',
            query_op in ('$lt', '$lte', '$gt', '$gte'))


def _get_ids_from_idx(idx, query_ops):  # noqa: C901
    """
    Returns the ids that match a set of query_ops in an index.

    :param idx SortedDict:
    :param query_ops str|dict:
    :rtype: set
    """
    if not isinstance(query_ops, dict):
        return set(idx.get(_make_idx_key(query_ops), set()))

    if not set(query_ops.keys()).intersection(_SUPPORTED_FILTER_OPERATORS):
        if _make_idx_key(query_ops) in idx.keys():
            return idx[_make_idx_key(query_ops)]
        return set()

    keys_remain = set(idx.keys())
    keys_not_cursed = keys_remain.copy()
    keys_cursed = set()

    for query_op, query_val in sorted(query_ops.items(),
                                      key=_idx_filter_sort, reverse=True):
        clean_idx_key = _make_idx_key(query_val)
        if query_op == '$eq':
            keys_remain = {clean_idx_key} if clean_idx_key in keys_remain else set()

        elif query_op == '$ne':
            _keys_cursed = set(k for k in keys_not_cursed if k == clean_idx_key)
            keys_remain -= _keys_cursed
            keys_not_cursed -= _keys_cursed
            keys_cursed.update(_keys_cursed)
        elif query_op == '$lt':
            keys_remain = _ids_given_irange_filters(keys_remain, idx,
                                                    maximum=clean_idx_key,
                                                    inclusive=(False, False))
        elif query_op == '$lte':
            keys_remain = _ids_given_irange_filters(keys_remain, idx,
                                                    maximum=clean_idx_key,
                                                    inclusive=(False, True))
        elif query_op == '$gt':
            keys_remain = _ids_given_irange_filters(keys_remain, idx,
                                                    minimum=clean_idx_key,
                                                    inclusive=(False, False))
        elif query_op == '$gte':
            keys_remain = _ids_given_irange_filters(keys_remain, idx,
                                                    minimum=clean_idx_key,
                                                    inclusive=(True, False))
        elif query_op == '$in':
            if not isinstance(query_val, (list, tuple, set)):
                raise MongitaError("'$in' requires an iterable")
            clean_q_val = [_make_idx_key(e) for e in query_val]
            keys_remain = set(k for k in keys_remain
                              if k in clean_q_val)
        elif query_op == '$nin':
            if not isinstance(query_val, (list, tuple, set)):
                raise MongitaError("'$nin' requires an iterable")
            clean_q_val = [_make_idx_key(e) for e in query_val]
            _keys_cursed = set(k for k in keys_not_cursed
                               if k in clean_q_val)
            keys_remain -= _keys_cursed
            keys_not_cursed -= _keys_cursed
            keys_cursed.update(_keys_cursed)
        # validation of options is done earlier
    ids_cursed = set()
    for k in keys_cursed:
        ids_cursed.update(idx[k])

    ret = set()
    for k in keys_remain:
        ret.update(idx[k])
    ret -= ids_cursed
    return ret


def _failed_update_error(update_op, update_op_dict, doc, msg):
    """Helper for raising errors on update"""
    return MongitaError("Cannot apply operation %r to %r (%s)" %
                        ({update_op: update_op_dict}, doc, msg))


def _update_item_in_doc(update_op, update_op_dict, doc):  # noqa: C901
    """
    Given an $update_op, a {doc_key: value} update_op_dict, and a doc,
    Update the doc in-place at doc_key with the update operation.

    e.g.
    doc = {'hi': 'ma'}
    update_op = '$set'
    update_op_dict {'ma': 'pa'}
    -> {'hi': 'pa'}

    :param update_op str:
    :param update_op_dict {str: value}:
    :param doc dict:
    :rtype: None
    """

    for doc_key, value in update_op_dict.items():
        ds, last_key = _get_datastructure_from_doc(doc, doc_key)
        if isinstance(ds, list):
            _rightpad(ds, last_key)
        if ds is None:
            raise _failed_update_error(update_op, update_op_dict, doc,
                                       "Could not find item")
        if update_op == '$set':
            ds[last_key] = value
        elif update_op == '$inc':
            if not isinstance(value, (int, float)):
                raise _failed_update_error(update_op, update_op_dict, doc,
                                           "Increment was not numeric")
            elif not isinstance(ds.get(last_key), (int, float)):
                raise _failed_update_error(update_op, update_op_dict, doc,
                                           "Document value was not numeric")
            ds[last_key] += value
        elif update_op == '$push':
            if isinstance(ds.get(last_key), list):
                ds[last_key].append(value)
            elif last_key not in ds:
                ds[last_key] = [value]
            else:
                raise _failed_update_error(update_op, update_op_dict, doc,
                                           "Document value was not a list")
        # Should never get an update key we don't recognize b/c _validate_update


def _rightpad(item, desired_length):
    """
    Given a list, pad to the desired_length with Nones
    This might be slow but it modifies the list in place

    :param item list:
    :param desired_length int:
    :rtype: None
    """
    pad_len = desired_length - len(item) + 1
    for _ in range(pad_len):
        item.append(None)


def _get_datastructure_from_doc(doc, key):  # noqa: C901
    """
    Get a pass-by-reference data structure from the document so that we can
    update it in-place. This dives deep into the document with the key
    parameter which uses dot notation.

    e.g.
    doc = {'deep': {'nested': {'list': [1, 2, 3]}}}
    key = 'deep.nested.list.5'
    -> a reference to [1, 2, 3, None, None] and 5

    :param doc dict:
    :param key str:
    :returns: the datastructure and the final accessor
    :rtype: list|dict|None, value
    """

    if '.' not in key:
        return doc, key

    item = doc
    levels = key.split('.')
    levels, last_level = levels[:-1], levels[-1]
    for level in levels:
        if isinstance(item, list):
            try:
                level_int = int(level)
            except ValueError:
                return None, None
            if level_int < 0:
                return None, None
            try:
                item = item[level_int]
            except IndexError:
                _rightpad(item, level_int)
                item = item[level_int] or {}
        elif isinstance(item, dict):
            if level not in item or not isinstance(item[level], (list, dict)):
                item[level] = {}
            item = item[level]
        else:
            return None, None
    if isinstance(item, list):
        try:
            last_level = int(last_level)
        except ValueError:
            return None, None
    return item, last_level


def _get_item_from_doc(doc, key):
    """
    Get an item from the document given a key which might use dot notation.

    e.g.
    doc = {'deep': {'nested': {'list': ['a', 'b', 'c']}}}
    key = 'deep.nested.list.1'
    -> 'b'

    :param doc dict:
    :param key str:
    :rtype: value
    """
    if '.' in key:
        item = doc
        for level in key.split('.'):
            if isinstance(item, list):
                try:
                    level_int = int(level)
                except ValueError:
                    return None
                try:
                    item = item[level_int]
                except IndexError:
                    return None
            elif isinstance(item, dict):
                item = item.get(level, {})
            else:
                return None
        return item or None
    return doc.get(key)


def _make_idx_key(idx_key):
    """
    MongoDB is very liberal when it comes to what keys it can compare on.
    When we get something weird, it makes sense to just store it as a
    hashable key

    :param idx_key value:
    :rtype: hashable value
    """
    if isinstance(idx_key, collections.abc.Hashable):
        return _sort_tup(idx_key)
    try:
        return _sort_tup(str(bson.encode(idx_key)))
    except TypeError:
        return _sort_tup(str(bson.encode({'idx_key': idx_key})))


def _update_idx_doc_with_new_documents(documents, idx_doc):
    """
    Update an idx_doc given documents which were just inserted / modified / etc

    :param documents List[dict]:
    :param idx_doc {key_str: str, direction: int idx: SortedDict, ...}:
    :rtype: None
    """
    documents = list(documents)
    _remove_docs_from_idx_doc(set(d['_id'] for d in documents), idx_doc)

    key_str = idx_doc['key_str']
    new_idx = sortedcontainers.SortedDict(idx_doc['idx'])

    for doc in documents:
        item_from_doc = _get_item_from_doc(doc, key_str)
        if isinstance(item_from_doc, list):
            for item in item_from_doc:
                key = _make_idx_key(item)
                new_idx.setdefault(key, set()).add(doc['_id'])
        key = _make_idx_key(item_from_doc)
        new_idx.setdefault(key, set()).add(doc['_id'])

    reverse = idx_doc['direction'] == DESCENDING
    idx_doc['idx'] = sortedcontainers.SortedDict(sorted(new_idx.items(), reverse=reverse))


def _remove_docs_from_idx_doc(doc_ids, idx_doc):
    """
    Update an idx_doc given documents which were just removed

    :param doc_ids set[str]:
    :param idx_doc {key_str: str, direction: int idx: SortedDict, ...}:
    :rtype: None
    """

    idx_doc_idx = idx_doc['idx']
    for k in idx_doc_idx.keys():
        idx_doc_idx[k] -= doc_ids


def _sort_tup(item):
    """
    Get sort tuple of item type according to mongodb rules

    :param item Value:
    :rtype: (int, Value)
    """
    try:
        return (SORT_ORDER[type(item)], item)
    except KeyError:
        pass
    # this assumes the item is None but could catch other
    # types if we are not careful. Sorting bugs are minor though
    return (b'\x01', item)


def _sort_func(doc, sort_key):
    """
    Sorter to sort different types according to MongoDB rules

    :param doc dict:
    :param sort_key str:
    :rtype: tuple
    """
    item = _get_item_from_doc(doc, sort_key)
    return _sort_tup(item)


def _sort_docs(docs, sort_list):
    """
    Given the sort list provided in the .sort() method,
    sort the documents in place.

    from https://docs.python.org/3/howto/sorting.html

    :param docs List[dict]:
    :param sort_list List[(key, direction)]
    :rtype: None
    """
    for sort_key, direction in reversed(sort_list):
        _sort_func_partial = functools.partial(_sort_func, sort_key=sort_key)
        if direction == ASCENDING:
            docs.sort(key=_sort_func_partial)
        elif direction == DESCENDING:
            docs.sort(key=_sort_func_partial, reverse=True)
        # validation on direction happens in cursor


def _split_filter(filter, metadata):
    """
    Split the filter into indx_ops and slow_filters which are later used
    differently

    :param filter {doc_key: query_ops}:
    :param metadata dict:
    :rtype: {doc_key: query_ops}, [(SortedDict idx, dict query_ops), ...]
    """
    slow_filters = {}
    indx_ops = []
    indexes = metadata.get('indexes', {})
    for doc_key, query_ops in filter.items():
        if doc_key + '_1' in indexes:
            indx_ops.append((indexes[doc_key + '_1']['idx'], query_ops))
        elif doc_key + '_-1' in indexes:
            indx_ops.append((indexes[doc_key + '_-1']['idx'], query_ops))
        else:
            slow_filters[doc_key] = query_ops
    return slow_filters, indx_ops


def _apply_indx_ops(indx_ops):
    """
    Return all doc_ids that can be found through the index filters

    :param indx_ops {idx_key: query_ops}:
    :param indexes dict:
    :rtype: set
    """
    doc_ids_so_far = set()
    for idx, query_ops in indx_ops:
        doc_ids = _get_ids_from_idx(idx, query_ops)
        if not doc_ids:
            return set()
        if doc_ids_so_far:
            doc_ids_so_far = doc_ids_so_far.intersection(doc_ids)
            if not doc_ids_so_far:
                return set()
        else:
            doc_ids_so_far = doc_ids
    return doc_ids_so_far


class Collection():
    UNIMPLEMENTED = ['aggregate', 'aggregate_raw_batches', 'bulk_write', 'codec_options',
                     'create_indexes', 'drop', 'drop_indexes', 'ensure_index',
                     'estimated_document_count', 'find_one_and_delete',
                     'find_one_and_replace', 'find_one_and_update', 'find_raw_batches',
                     'inline_map_reduce', 'list_indexes', 'map_reduce', 'next',
                     'options', 'read_concern', 'read_preference', 'rename', 'watch', ]
    DEPRECATED = ['reindex', 'parallel_scan', 'initialize_unordered_bulk_op',
                  'initialize_ordered_bulk_op', 'group', 'count', 'insert', 'save',
                  'update', 'remove', 'find_and_modify', 'ensure_index']

    def __init__(self, collection_name, database, write_concern=None, read_concern=None):
        self.name = collection_name
        self.database = database
        self._write_concern = write_concern or WriteConcern()
        self._read_concern = read_concern or ReadConcern()
        self._engine = database._engine
        self._existence_verified = False
        self._base_location = f'{database.name}.{collection_name}'

    def __repr__(self):
        return "Collection(%s, %r)" % (repr(self.database), self.name)

    def __getattr__(self, attr):
        """
        First check for deprecated / unimplemented.
        Then, MongoDB has this weird thing where there can be dots in a collection
        name.
        """
        if attr in self.DEPRECATED:
            raise MongitaNotImplementedError.create_depr("Collection", attr)
        if attr in self.UNIMPLEMENTED:
            raise MongitaNotImplementedError.create("Collection", attr)
        return Collection(collection_name=self.name + '.' + attr,
                          database=self.database)

    @property
    def full_name(self):
        return self._base_location

    @property
    def write_concern(self):
        return self._write_concern

    @property
    def read_concern(self):
        return self._read_concern

    def with_options(self, **kwargs):
        write_concern = kwargs.pop('write_concern', None)
        read_concern = kwargs.pop('read_concern', None)

        if kwargs:
            raise MongitaNotImplementedError("The method 'with_options' doesn't yet "
                                             "accept %r" % kwargs)
        return Collection(self.name, self.database,
                          write_concern=write_concern,
                          read_concern=read_concern)

    def __create(self):
        """
        MongoDB doesn't require you to explicitly create collections. They
        are created when first accessed. This creates the collection and is
        called early in modifier methods.
        """
        if self._existence_verified:
            return
        with self._engine.lock:
            if not self._engine.get_metadata(self._base_location):
                self._engine.create_path(self._base_location)
                metadata = MetaStorageObject(copy.deepcopy(_DEFAULT_METADATA))
                assert self._engine.put_metadata(self._base_location, metadata)
            self.database.__create(self.name)
        self._existence_verified = True

    def __insert_one(self, document):
        """
        Insert a single document.

        :param document dict:
        :rtype: None
        """
        success = self._engine.put_doc(self.full_name, document,
                                       no_overwrite=True)
        if not success:
            assert self._engine.doc_exists(self.full_name, document['_id'])
            raise DuplicateKeyError("Document %r already exists" % document['_id'])

    @support_alert
    async def insert_one(self, document):
        """
        Insert a single document.

        :param document dict:
        :rtype: results.InsertOneResult
        """
        _validate_doc(document)
        document = copy.deepcopy(document)
        document['_id'] = document.get('_id') or bson.ObjectId()
        self.__create()
        with self._engine.lock:
            metadata = self.__get_metadata()
            self.__insert_one(document)
            self.__update_indicies([document], metadata)
        return InsertOneResult(document['_id'])

    @support_alert
    async def insert_many(self, documents, ordered=True):
        """
        Insert documents. If ordered, stop inserting if there is an error.
        If not ordered, all operations are attempted

        :param list documents:
        :param bool ordered:
        :rtype: results.InsertManyResult
        """
        if not isinstance(documents, list):
            raise MongitaError("Documents must be a list")
        ready_docs = []
        for doc in documents:
            _validate_doc(doc)
            doc = copy.deepcopy(doc)
            doc['_id'] = doc.get('_id') or bson.ObjectId()
            ready_docs.append(doc)
        self.__create()
        success_docs = []
        exception = None
        with self._engine.lock:
            metadata = self.__get_metadata()
            for doc in ready_docs:
                try:
                    self.__insert_one(doc)
                    success_docs.append(doc)
                except Exception as ex:  # pylint: disable=broad-except
                    if ordered:
                        self.__update_indicies(success_docs, metadata)
                        raise MongitaError("Ending insert_many because of error") from ex
                    exception = ex
                    continue
            self.__update_indicies(success_docs, metadata)
        if exception:
            raise MongitaError("Not all documents inserted") from exception
        return InsertManyResult(success_docs)

    @support_alert
    async def replace_one(self, filter, replacement, upsert=False):
        """
        Replace one document. If no document was found with the filter,
        and upsert is True, insert the replacement.

        :param filter dict:
        :param replacement dict:
        :param bool upsert:
        :rtype: results.UpdateResult
        """
        filter = filter or {}
        _validate_filter(filter)
        _validate_doc(replacement)
        self.__create()

        replacement = copy.deepcopy(replacement)

        with self._engine.lock:
            doc_id = self.__find_one_id(filter, upsert=upsert)
            if not doc_id:
                if upsert:
                    metadata = self.__get_metadata()
                    replacement['_id'] = replacement.get('_id') or bson.ObjectId()
                    self.__insert_one(replacement)
                    self.__update_indicies([replacement], metadata)
                    return UpdateResult(0, 1, replacement['_id'])
                return UpdateResult(0, 0)
            replacement['_id'] = doc_id
            metadata = self.__get_metadata()
            assert self._engine.put_doc(self.full_name, replacement)
            self.__update_indicies([replacement], metadata)
            return UpdateResult(1, 1)

    def __find_one_id(self, filter, sort=None, skip=None, upsert=False):
        """
        Given the filter, return a single object_id or None.

        :param filter dict:
        :param sort list[(key, direction)]|None
        :param skip int|None
        :rtype: str|None
        """

        if not filter and not sort:
            return self._engine.find_one_id(self._base_location)

        if '_id' in filter:
            if upsert or self._engine.doc_exists(self.full_name, filter['_id']):
                return filter['_id']
            return None

        try:
            return next(self.__find_ids(filter, sort, skip=skip))
        except StopIteration:
            return None

    def __find_one(self, filter, sort, skip):
        """
        Given the filter, return a single doc or None.

        :param filter dict:
        :param sort list[(key, direction)]|None
        :param skip int|None
        :rtype: dict|None
        """
        doc_id = self.__find_one_id(filter, sort, skip)
        if doc_id:
            doc = self._engine.get_doc(self.full_name, doc_id)
            if doc:
                return copy.deepcopy(doc)

    def __find_ids(self, filter, sort=None, limit=None, skip=None, metadata=None):  # noqa: C901
        """
        Given a filter, find all doc_ids that match this filter.
        Be sure to also sort and limit them.
        This method will download documents for non-indexed filters (slow_filters).
        Downloaded docs are cached in the engine layer so performance cost is minimal.
        This method returns a generator

        :param filter dict:
        :param sort list[(key, direction)]|None:
        :param limit int|None:
        :param skip int|None:
        :param metadata dict|None:
        :rtype: Generator(list[str])
        """
        filter = filter or {}
        sort = sort or []

        if limit == 0:
            return

        metadata = metadata or self.__get_metadata()
        slow_filters, indx_ops = _split_filter(filter, metadata)

        # If we have index ops, we can use those ids as a starting point.
        # otherwise, we need to get all_ids and filter one-by-one
        if indx_ops:
            doc_ids = _apply_indx_ops(indx_ops)
        else:
            doc_ids = self._engine.list_ids(self._base_location)
        if not doc_ids:
            return

        if sort:
            docs_to_return = []
            for doc_id in doc_ids:
                doc = self._engine.get_doc(self.full_name, doc_id)
                if _doc_matches_slow_filters(doc, slow_filters):
                    docs_to_return.append(doc)
            _sort_docs(docs_to_return, sort)

            if skip:
                docs_to_return = docs_to_return[skip:]

            if limit is None:
                for doc in docs_to_return:
                    yield doc['_id']
            else:
                i = 0
                for doc in docs_to_return:
                    yield doc['_id']
                    i += 1
                    if i == limit:
                        return
            return

        if skip:
            doc_ids = doc_ids[skip:]

        if limit is None:
            for doc_id in doc_ids:
                doc = self._engine.get_doc(self.full_name, doc_id)
                if doc and _doc_matches_slow_filters(doc, slow_filters):
                    yield doc['_id']
            return

        i = 0
        for doc_id in doc_ids:
            doc = self._engine.get_doc(self.full_name, doc_id)
            if _doc_matches_slow_filters(doc, slow_filters):
                yield doc['_id']
                i += 1
                if i == limit:
                    return

    def __find(self, filter, sort=None, limit=None, skip=None, metadata=None, shallow=False):
        """
        Given a filter, find all docs that match this filter.
        This method returns a generator.

        :param filter dict:
        :param sort list[(key, direction)]|None:
        :param limit int|None:
        :param skip int|None:
        :param metadata dict|None:
        :rtype: Generator(list[dict])
        """
        gen = self.__find_ids(filter, sort, limit, skip, metadata=metadata)

        if shallow:
            for doc_id in gen:
                doc = self._engine.get_doc(self.full_name, doc_id)
                yield doc
        else:
            for doc_id in gen:
                doc = self._engine.get_doc(self.full_name, doc_id)
                yield copy.deepcopy(doc)

    @support_alert
    async def find_one(self, filter=None, sort=None, skip=None):
        """
        Return the first matching document.

        :param filter dict:
        :param sort list[(key, direction)]|None:
        :param skip int|None:
        :rtype: dict|None
        """

        filter = filter or {}
        _validate_filter(filter)

        if sort is not None:
            sort = _validate_sort(sort)
        return self.__find_one(filter, sort, skip)

    @support_alert
    def find(self, filter=None, sort=None, limit=None, skip=None):
        """
        Return a cursor of all matching documents.

        :param filter dict:
        :param sort list[(key, direction)]|None:
        :param limit int|None:
        :param skip int|None:
        :rtype: cursor.Cursor
        """

        filter = filter or {}
        _validate_filter(filter)

        if sort is not None:
            sort = _validate_sort(sort)

        if limit is not None and not isinstance(limit, int):
            raise TypeError('Limit must be an integer')

        if skip is not None:
            if not isinstance(skip, int):
                raise TypeError('Skip must be an integer')
            if skip < 0:
                raise ValueError('Skip must be >=0')

        return Cursor(self.__find, filter, sort, limit, skip)

    def __update_doc(self, doc_id, update):
        """
        Given a doc_id and an update dict, find the document and safely update it.
        Returns the updated document

        :param doc_id str:
        :param update dict:
        :rtype: dict
        """
        doc = self._engine.get_doc(self.full_name, doc_id)
        for update_op, update_op_dict in update.items():
            _update_item_in_doc(update_op, update_op_dict, doc)
        assert self._engine.put_doc(self.full_name, doc)
        return dict(doc)

    @support_alert
    async def update_one(self, filter, update, upsert=False):
        """
        Find one document matching the filter and update it.
        The 'upsert' parameter is not supported.

        :param filter dict:
        :param update dict:
        :param upsert bool:
        :rtype: results.UpdateResult
        """

        _validate_filter(filter)
        _validate_update(update)
        self.__create()
        if upsert:
            raise MongitaNotImplementedError("Mongita does not support 'upsert' on "
                                             "update operations. Use `replace_one`.")

        with self._engine.lock:
            doc_ids = list(self.__find_ids(filter))
            matched_count = len(doc_ids)
            if not matched_count:
                return UpdateResult(matched_count, 0)
            metadata = self.__get_metadata()
            doc = self.__update_doc(doc_ids[0], update)
            self.__update_indicies([doc], metadata)
        return UpdateResult(matched_count, 1)

    @support_alert
    async def update_many(self, filter, update, upsert=False):
        """
        Update every document matched by the filter.
        The 'upsert' parameter is not supported.

        :param filter dict:
        :param update dict:
        :param upsert bool:
        :rtype: results.UpdateResult
        """
        _validate_filter(filter)
        _validate_update(update)
        self.__create()
        if upsert:
            raise MongitaNotImplementedError("Mongita does not support 'upsert' "
                                             "on update operations. Use `replace_one`.")

        success_docs = []
        matched_cnt = 0
        with self._engine.lock:
            doc_ids = list(self.__find_ids(filter))
            metadata = self.__get_metadata()
            for doc_id in doc_ids:
                doc = self.__update_doc(doc_id, update)
                success_docs.append(doc)
                matched_cnt += 1
            self.__update_indicies(success_docs, metadata)
        return UpdateResult(matched_cnt, len(success_docs))

    @support_alert
    async def delete_one(self, filter):
        """
        Delete one document matching the filter.

        :param filter dict:
        :rtype: results.DeleteResult
        """
        _validate_filter(filter)
        self.__create()

        with self._engine.lock:
            doc_id = self.__find_one_id(filter)
            if not doc_id:
                return DeleteResult(0)
            metadata = self.__get_metadata()
            self._engine.delete_doc(self.full_name, doc_id)
            self.__update_indicies_deletes({doc_id}, metadata)
        return DeleteResult(1)

    @support_alert
    async def delete_many(self, filter):
        """
        Delete all documents matching the filter.

        :param filter dict:
        :rtype: results.DeleteResult
        """
        _validate_filter(filter)
        self.__create()

        success_deletes = set()
        with self._engine.lock:
            doc_ids = list(self.__find_ids(filter))
            metadata = self.__get_metadata()
            for doc_id in doc_ids:
                if self._engine.delete_doc(self.full_name, doc_id):
                    success_deletes.add(doc_id)
            self.__update_indicies_deletes(success_deletes, metadata)
        return DeleteResult(len(success_deletes))

    @support_alert
    async def count_documents(self, filter):
        """
        Returns a count of all documents matching the filter.
        This can be much faster than taking the length of a find query.

        :param filter dict:
        :rtype: int
        """
        _validate_filter(filter)
        return len(list(self.__find_ids(filter)))

    @support_alert
    async def distinct(self, key, filter=None):
        """
        Given a key, return all distinct documents matching the key

        :param key str:
        :param filter dict|None:
        :rtype: list[str]
        """

        if not isinstance(key, str):
            raise MongitaError("The 'key' parameter must be a string")
        filter = filter or {}
        _validate_filter(filter)
        uniq = set()
        for doc in self.find(filter):
            uniq.add(_get_item_from_doc(doc, key))
        uniq.discard(None)
        return list(uniq)

    def __get_metadata(self):
        """
        Thin wrapper to get metadata.
        Always be sure to lock the engine when modifying metadata

        :rtype: dict
        """
        return self._engine.get_metadata(self._base_location) or copy.deepcopy(_DEFAULT_METADATA)

    def __update_indicies_deletes(self, doc_ids, metadata):
        """
        Given a list of deleted document ids, remove those documents from all indexes.
        Returns the new metadata dictionary.

        :param doc_ids set[str]:
        :param metadata dict:
        :rtype: dict
        """
        if not doc_ids:
            return metadata
        for idx_doc in metadata.get('indexes', {}).values():
            _remove_docs_from_idx_doc(doc_ids, idx_doc)
        assert self._engine.put_metadata(self._base_location, metadata)
        return metadata

    def __update_indicies(self, documents, metadata):
        """
        Given a list of deleted document ids, add those documents to all indexes.
        Returns the new metadata dictionary.

        :param documents list[dict]:
        :param metadata dict:
        :rtype: dict
        """
        for idx_doc in metadata.get('indexes', {}).values():
            _update_idx_doc_with_new_documents(documents, idx_doc)
        assert self._engine.put_metadata(self._base_location, metadata)
        return metadata

    @support_alert
    async def create_index(self, keys, background=False):
        """
        Create a new index for the collection.
        Indexes can dramatically speed up queries that use its fields.
        Currently, only single key indicies are supported.
        Returns the name of the new index.

        :param keys str|[(key, direction)]:
        :param background bool:
        :rtype: str
        """
        if background is not False:
            raise MongitaNotImplementedError("background index creation is not supported")

        self.__create()

        if isinstance(keys, str):
            keys = [(keys, ASCENDING)]
        if not isinstance(keys, list) or keys == []:
            raise MongitaError("Unsupported keys parameter format %r. "
                               "See the docs." % str(keys))
        if len(keys) > 1:
            raise MongitaNotImplementedError("Mongita does not support multi-key indexes yet")
        for k, direction in keys:
            if not k or not isinstance(k, str):
                raise MongitaError("Index keys must be strings %r" % str(k))
            if direction not in (ASCENDING, DESCENDING):
                raise MongitaError("Index key direction must be either ASCENDING (1) "
                                   "or DESCENDING (-1). Not %r" % direction)

        key_str, direction = keys[0]

        idx_name = f'{key_str}_{direction}'
        new_idx_doc = {
            '_id': idx_name,
            'key_str': key_str,
            'direction': direction,
            'idx': {},
        }

        with self._engine.lock:
            metadata = self.__get_metadata()
            _update_idx_doc_with_new_documents(self.__find({}, metadata=metadata, shallow=True),
                                               new_idx_doc)
            metadata['indexes'][idx_name] = new_idx_doc
            assert self._engine.put_metadata(self._base_location, metadata)
        return idx_name

    @support_alert
    async def drop_index(self, index_or_name):
        """
        Drops the index given by the index_or_name parameter. Passing index
        objects is not supported

        :param index_or_name str|(str, int)
        :rtype: None
        """
        self.__create()

        if isinstance(index_or_name, (list, tuple)) \
                and len(index_or_name) == 1 \
                and len(index_or_name[0]) == 2 \
                and isinstance(index_or_name[0][0], str) \
                and isinstance(index_or_name[0][1], int):
            index_or_name = f'{index_or_name[0][0]}_{index_or_name[0][1]}'
        if not isinstance(index_or_name, str):
            raise MongitaError("Unsupported index_or_name parameter format. See the docs.")
        if re.match(r'^.*?_\-?1$', index_or_name):
            _, direction = index_or_name.rsplit('_', 1)
            direction = int(direction)
        else:
            index_or_name = index_or_name + '_1'

        with self._engine.lock:
            metadata = self.__get_metadata()
            if index_or_name not in metadata['indexes']:
                raise OperationFailure('Index not found with name %r' % index_or_name)
            del metadata['indexes'][index_or_name]
            assert self._engine.put_metadata(self._base_location, metadata)

    @support_alert
    async def index_information(self):
        """
        Returns a list of indexes in the collection

        :rtype: {idx_id: {'key': [(key_str, direction_int)]}}
        """

        ret = [{'_id_': {'key': [('_id', 1)]}}]
        metadata = self.__get_metadata()
        for idx in metadata.get('indexes', {}).values():
            ret.append({idx['_id']: {'key': [(idx['key_str'], idx['direction'])]}})
        return ret



from typing import List, Tuple, Literal

from retk.models.client import client
from retk.models.tps import UserFile


def get_files(
        uid: str,
        page: int,
        page_size: int,
        sort: Literal["filename", "created_at", "size"] = "created_at",
) -> Tuple[List[UserFile], int]:
    docs = client.coll.user_file.find({"uid": uid})
    total = docs.count()
    if sort == "filename":
        docs = docs.sort([("filename", 1), ("_id", -1)])
    elif sort == "created_at":
        docs = docs.sort("_id", direction=-1)
    elif sort == "size":
        docs = docs.sort([("size", -1), ("_id", -1)])

    docs = docs.skip((page - 1) * page_size).limit(page_size)
    return list(docs), total



from pydantic import BaseModel, EmailStr, Field

from retk.const import settings, LanguageEnum


class TokenResponse(BaseModel):
    requestId: str
    token: str = ""


class SignupRequest(BaseModel):
    email: EmailStr = Field(max_length=settings.EMAIL_MAX_LENGTH)
    password: str = Field(max_length=settings.PASSWORD_MAX_LENGTH)
    verificationToken: str = Field(max_length=2000)
    verification: str = Field(max_length=10)
    language: LanguageEnum


class LoginRequest(BaseModel):
    email: EmailStr = Field(max_length=settings.EMAIL_MAX_LENGTH)
    password: str = Field(max_length=settings.PASSWORD_MAX_LENGTH)
    language: LanguageEnum = LanguageEnum.EN


class EmailVerificationRequest(BaseModel):
    email: str = Field(max_length=settings.EMAIL_MAX_LENGTH)
    userExistOk: bool = Field(default=False)
    captchaToken: str = Field(max_length=2000)
    captchaCode: str = Field(max_length=10)
    language: LanguageEnum = LanguageEnum.EN


class ForgetPasswordRequest(BaseModel):
    email: str = Field(max_length=settings.EMAIL_MAX_LENGTH)
    newPassword: str = Field(max_length=settings.PASSWORD_MAX_LENGTH)
    verification: str = Field(max_length=10)
    verificationToken: str = Field(max_length=2000)
    language: LanguageEnum = LanguageEnum.EN



from . import tasks
from . import timing
from .schedule import (  # noqa: F401
    init_tasks,
    start,
    stop,
    get_jobs,
    get_job,
    clear_jobs,
    run_once_at,
    run_once_after,
    run_once_now,
    run_every_at,
)



from enum import Enum, unique


@unique
class LanguageEnum(str, Enum):
    EN = "en"
    ZH = "zh"

    def __str__(self):
        return self.value

    @classmethod
    def from_str(cls, s: str):
        if s == "en":
            return cls.EN
        elif s == "zh":
            return cls.ZH
        else:
            raise ValueError(f"Invalid language: {s}")

    @classmethod
    def is_valid(cls, item: str):
        return item in [v.value for v in cls.__members__.values()]



import datetime
import time
import unittest

from bson import ObjectId

from retk import const
from retk.models.search_engine.engine_local import LocalSearcher, SearchDoc
from retk.models.tps import AuthedUser
from . import utils


class LocalSearchTest(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.local")
        cls.searcher = LocalSearcher()

    @classmethod
    def tearDownClass(cls) -> None:
        utils.drop_env(".env.test.local")

    async def asyncSetUp(self) -> None:
        await self.searcher.drop()
        await self.searcher.init()
        self.assertTrue(self.searcher.index_path.exists())
        self.au = AuthedUser(
            u=AuthedUser.User(
                _id=ObjectId(),
                id="uid",
                source=0,
                account="rethink",
                nickname="rethink",
                email="rethink@rethink.run",
                avatar="",
                hashed="",
                disabled=False,
                modified_at=datetime.datetime.now(),
                used_space=0,
                type=0,

                last_state=AuthedUser.User.LastState(
                    node_display_method=0,
                    node_display_sort_key="",
                    recent_search=[],
                    recent_cursor_search_selected_nids=[],
                ),
                settings=AuthedUser.User.Settings(
                    language="en",
                    theme="light",
                    editor_mode="markdown",
                    editor_font_size=14,
                    editor_code_theme="github",
                    editor_sep_right_width=0,
                    editor_side_current_tool_id="",
                ),
            ),
            language="en",
            request_id="request_id",
        )

    async def asyncTearDown(self) -> None:
        await self.searcher.drop()

    async def test_add(self):
        for i in range(20):
            code = await self.searcher.add(au=self.au, doc=SearchDoc(
                nid=f"nid{i}",
                title=f"title {i}",
                body=f"this is {i} doc, 这是第 {i} 个文档",
            ))
            self.assertEqual(const.CodeEnum.OK, code)
            time.sleep(0.0001)

        docs, total = await self.searcher.search(
            au=self.au,
            query="title doc",
            sort_key="createdAt",
            reverse=True,
            page=0,
            limit=10,
        )

        self.assertEqual(10, len(docs))
        self.assertEqual(20, total)
        self.assertEqual("nid19", docs[0].nid)
        self.assertEqual("nid10", docs[-1].nid)
        self.assertEqual(['19 <em class="match term1">doc</em>, 这是第 19 个文档'], docs[0].bodyHighlights)
        self.assertEqual("<em class=\"match term0\">title</em> 19", docs[0].titleHighlight)

        docs, total = await self.searcher.search(
            au=self.au,
            query="doc",
            page=200,
            limit=10,
        )
        self.assertEqual(0, len(docs))
        self.assertEqual(20, total)

        docs, total = await self.searcher.search(
            au=self.au,
            query="",
            page=0,
            limit=50,
        )
        self.assertEqual(20, len(docs))
        self.assertEqual(20, total)

        count = await self.searcher.count_all()
        self.assertEqual(20, count)

    async def test_batch_add_update_delete(self):
        code = await self.searcher.add_batch(au=self.au, docs=[
            SearchDoc(
                nid=f"nid{i}",
                title=f"title{i}",
                body=f"this is {i} doc, 这是第 {i} 个文档",
            ) for i in range(20)
        ])
        self.assertEqual(const.CodeEnum.OK, code)

        docs, total = await self.searcher.search(
            au=self.au,
            query="doc",
            sort_key="createdAt",
            reverse=True,
            page=0,
            limit=10,
        )
        self.assertEqual(10, len(docs))
        self.assertEqual(20, total)

        code = await self.searcher.update_batch(au=self.au, docs=[
            SearchDoc(
                nid=f"nid{i}",
                title=f"title_update{i}",
                body=f"this is {i} doc, 这是第 {i} 个文档",
            ) for i in range(20)
        ])
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(20, await self.searcher.count_all())

        docs, total = await self.searcher.search(
            au=self.au,
            query="doc",
            sort_key="createdAt",
            reverse=True,
            page=0,
            limit=10,
        )
        self.assertEqual(10, len(docs))
        self.assertEqual(20, total)
        self.assertEqual("nid19", docs[0].nid)

        code = await self.searcher.disable(au=self.au, nid="nid18")
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(20, await self.searcher.count_all())

        docs, total = await self.searcher.search(
            au=self.au,
            query="doc",
            sort_key="createdAt",
            reverse=True,
            page=0,
            limit=10,
        )
        self.assertEqual(10, len(docs))
        self.assertEqual(19, total)
        self.assertEqual("nid17", docs[1].nid)
        self.assertEqual(20, await self.searcher.count_all())

        code = await self.searcher.enable(au=self.au, nid="nid18")
        self.assertEqual(const.CodeEnum.OK, code)

        code = await self.searcher.batch_to_trash(au=self.au, nids=[f"nid{i}" for i in range(10)])
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(20, await self.searcher.count_all())
        code = await self.searcher.delete_batch(au=self.au, nids=[f"nid{i}" for i in range(10)])
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(10, await self.searcher.count_all())

        docs, total = await self.searcher.search(
            au=self.au,
            query="doc",
            sort_key="createdAt",
            reverse=True,
            page=0,
            limit=10,
        )
        self.assertEqual(10, len(docs))
        self.assertEqual(10, total)
        self.assertEqual("nid19", docs[0].nid)
        self.assertEqual("nid18", docs[1].nid)
        self.assertEqual("nid17", docs[2].nid)

        docs, total = await self.searcher.search(
            au=self.au,
            query="doc",
            sort_key="createdAt",
            reverse=True,
            page=0,
            limit=10,
            exclude_nids=["nid19"]
        )
        self.assertEqual(9, len(docs))
        self.assertEqual(9, total)
        self.assertEqual("nid18", docs[0].nid)

        docs, total = await self.searcher.search(
            au=self.au,
            query="doc",
            sort_key="title",
            reverse=False,
            page=0,
            limit=10,
            exclude_nids=["nid19"]
        )
        self.assertEqual(9, len(docs))
        self.assertEqual(9, total)
        self.assertEqual("nid10", docs[0].nid)



import datetime
import time
import unittest
from copy import deepcopy
from textwrap import dedent
from unittest.mock import patch

import elastic_transport
import pymongo.errors
from bson import ObjectId

from retk import const, config, core
from retk.controllers.schemas.user import PatchUserRequest
from retk.core.account.manager import signup
from retk.core.scheduler import tasks
from retk.models import db_ops
from retk.models.client import client
from retk.models.tps import AuthedUser, convert_user_dict_to_authed_user
from retk.utils import get_token
from . import utils


class RemoteModelsTest(unittest.IsolatedAsyncioTestCase):
    default_pwd = "rethink123"

    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.dev")

    @classmethod
    def tearDownClass(cls) -> None:
        utils.drop_env(".env.test.dev")

    async def asyncSetUp(self) -> None:
        if utils.skip_no_connect.skip:
            print("remote test asyncSetUp skipped")
            return

        try:
            client.connection_timeout = 1
            await client.init()
            for coll in client.coll.__dict__.values():
                await coll.delete_many({})

            u, code = await signup(
                email=const.DEFAULT_USER["email"],
                password=self.default_pwd,
                language=const.LanguageEnum.EN.value)
            self.assertEqual(const.CodeEnum.OK, code)
            self.access_token, self.refresh_token = get_token(
                uid=u["id"],
                language=const.LanguageEnum.EN.value,
            )

            self.au = AuthedUser(
                u=convert_user_dict_to_authed_user(u),
                request_id="xxx",
                language=u["settings"]["language"],
            )
        except (
                pymongo.errors.NetworkTimeout,
                pymongo.errors.ServerSelectionTimeoutError,
                elastic_transport.ConnectionError,
        ):
            try:
                await client.drop()
            except pymongo.errors.ServerSelectionTimeoutError:
                pass
            print("remote test asyncSetUp timeout")
            utils.skip_no_connect.skip = True

    async def asyncTearDown(self) -> None:
        if utils.skip_no_connect.skip:
            print("remote test asyncTearDown skipped")
            return
        try:
            await client.drop()
        except (
                pymongo.errors.NetworkTimeout,
                pymongo.errors.ServerSelectionTimeoutError,
                elastic_transport.ConnectionError,
                RuntimeError,
        ):
            print("remote test asyncTearDown timeout")
            utils.skip_no_connect.skip = True

    @utils.skip_no_connect
    async def test_same_key(self):
        async def add():
            oid = ObjectId()
            await client.coll.users.insert_one({
                "_id": oid,
                "id": "same",
                "account": "a",
                "source": 0,
                "email": "email",
                "hashed": "hashed",
                "avatar": "avatar",
                "disabled": False,
                "nickname": "nickname",
                "modifiedAt": oid.generation_time,
                "nodeIds": [],
                "usedSpace": 0,
                "type": const.USER_TYPE.NORMAL.id,
                "lastState": {
                    "recentCursorSearchSelectedNIds": [],
                    "recentSearch": [],
                    "nodeDisplayMethod": const.NodeDisplayMethodEnum.CARD.value,
                    "nodeDisplaySortKey": "modifiedAt"
                },
                "settings": {
                    "language": "en",
                    "editorMode": const.app.EditorModeEnum.WYSIWYG.value,
                    "editorTheme": const.app.AppThemeEnum.LIGHT.value,
                    "editorFontSize": 15,
                    "editorCodeTheme": const.app.EditorCodeThemeEnum.GITHUB.value,
                }
            })
            us = await client.coll.users.find({"id": "same"}).to_list(length=2)
            self.assertIsNotNone(us)
            self.assertEqual(1, len(us))
            self.assertEqual("a", us[0]["account"])

        await add()
        with self.assertRaises(pymongo.errors.DuplicateKeyError):
            await add()

    @utils.skip_no_connect
    async def test_user(self):
        u, code = await core.user.get_by_email(email=const.DEFAULT_USER["email"])
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual("rethink", u["nickname"])
        self.assertIsNotNone(u)

        u, code = await signup(email="aaa@rethink.run", password="bbb123")
        _id = u["id"]
        self.assertNotEqual("", _id)
        self.assertEqual(const.CodeEnum.OK, code)

        u, code = await core.user.get(_id)
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual("aaa", u["nickname"])

        u, code = await core.user.patch(
            au=AuthedUser(
                u=convert_user_dict_to_authed_user(u),
                request_id="xxx",
                language=const.LanguageEnum.EN.value,
            ),
            req=PatchUserRequest(
                nickname="2",
                avatar="3",
                lastState=PatchUserRequest.LastState(
                    nodeDisplayMethod=const.NodeDisplayMethodEnum.LIST.value,
                )
            )
        )
        self.assertEqual(const.CodeEnum.OK, code)

        u, code = await core.user.get(_id)
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual("2", u["nickname"])
        self.assertEqual("3", u["avatar"])

        code = await core.account.manager.disable_by_uid(uid=_id)
        self.assertEqual(const.CodeEnum.OK, code)

        u, code = await core.user.get(uid=_id, disabled=False)
        self.assertEqual(const.CodeEnum.USER_NOT_EXIST, code)

        u, code = await core.user.get(uid=_id, disabled=None)
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertTrue(u["disabled"])

        code = await core.account.manager.enable_by_uid(uid=_id)
        self.assertEqual(const.CodeEnum.OK, code)

        code = await core.account.manager.disable_by_uid(uid="ssaa")
        self.assertEqual(const.CodeEnum.USER_NOT_EXIST, code)

        await core.account.manager.delete_by_uid(uid=_id)

    @utils.skip_no_connect
    @patch("retk.core.node.backup.__remove_md_all_versions_from_cos")
    @patch("retk.core.node.backup.__remove_md_from_cos")
    @patch("retk.core.node.backup.__get_md_from_cos")
    @patch("retk.core.node.backup.__save_md_to_cos")
    async def test_node(
            self,
            mock_save_md_to_cos,
            mock_get_md_from_cos,
            mock_remove_md_from_cos,
            mock_remove_md_all_versions_from_cos,
    ):
        mock_save_md_to_cos.return_value = const.CodeEnum.OK
        mock_get_md_from_cos.return_value = ("", const.CodeEnum.OK)
        mock_remove_md_from_cos.return_value = const.CodeEnum.OK
        mock_remove_md_all_versions_from_cos.return_value = const.CodeEnum.OK

        u, code = await core.user.get(self.au.u.id)
        self.assertEqual(const.CodeEnum.OK, code)
        used_space = u["usedSpace"]
        node, code = await core.node.post(
            au=self.au, md="title\ntext", type_=const.NodeTypeEnum.MARKDOWN.value
        )
        self.assertEqual(const.CodeEnum.OK, code)
        u, code = await core.user.get(self.au.u.id)
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(used_space + len(node["md"].encode("utf-8")), u["usedSpace"])

        n, code = await core.node.get(au=self.au, nid=node["id"])
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual("title", n["title"])

        await client.search.refresh()
        ns, total = await client.search.search(au=self.au)
        self.assertEqual(3, len(ns))
        self.assertEqual(3, total)

        u, code = await core.user.get(self.au.u.id)
        self.assertEqual(const.CodeEnum.OK, code)
        used_space = u["usedSpace"]
        n, _, code = await core.node.update_md(au=self.au, nid=node["id"], md="# title2\ntext2")
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual("title2", n["title"])
        self.assertEqual("text2", n["snippet"])
        self.assertEqual(const.NodeTypeEnum.MARKDOWN.value, n["type"])

        u, code = await core.user.get(self.au.u.id)
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(used_space + (
                len(n["md"].encode("utf-8")) -
                len(node["md"].encode("utf-8"))
        ), u["usedSpace"])

        code = await core.node.disable(au=self.au, nid=node["id"])
        self.assertEqual(const.CodeEnum.OK, code)
        await client.search.refresh()
        n, code = await core.node.get(au=self.au, nid=node["id"])
        self.assertEqual(const.CodeEnum.NODE_NOT_EXIST, code)

        code = await core.node.to_trash(au=self.au, nid=node["id"])
        self.assertEqual(const.CodeEnum.OK, code)
        await client.search.refresh()
        code = await core.node.delete(au=self.au, nid=node["id"])
        self.assertEqual(const.CodeEnum.OK, code)
        n, code = await core.node.get(au=self.au, nid=node["id"])
        self.assertIsNone(n)
        self.assertEqual(const.CodeEnum.NODE_NOT_EXIST, code)

        u, code = await core.user.get(self.au.u.id)
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(used_space - len(node["md"].encode("utf-8")), u["usedSpace"])

        nodes, total = await core.node.core_nodes(au=self.au, page=0, limit=10)
        self.assertEqual(2, len(nodes))
        self.assertEqual(2, total)

    @utils.skip_no_connect
    @patch("retk.core.node.backup.__remove_md_all_versions_from_cos")
    @patch("retk.core.node.backup.__remove_md_from_cos")
    @patch("retk.core.node.backup.__get_md_from_cos")
    @patch("retk.core.node.backup.__save_md_to_cos")
    async def test_parse_at(
            self,
            mock_save_md_to_cos,
            mock_get_md_from_cos,
            mock_remove_md_from_cos,
            mock_remove_md_all_versions_from_cos,
    ):
        mock_save_md_to_cos.return_value = const.CodeEnum.OK
        mock_get_md_from_cos.return_value = ("", const.CodeEnum.OK)
        mock_remove_md_from_cos.return_value = const.CodeEnum.OK
        mock_remove_md_all_versions_from_cos.return_value = const.CodeEnum.OK

        nid1, _ = await core.node.post(
            au=self.au, md="c", type_=const.NodeTypeEnum.MARKDOWN.value,
        )
        nid2, _ = await core.node.post(
            au=self.au, md="我133", type_=const.NodeTypeEnum.MARKDOWN.value,
        )
        md = dedent(
            f"""title
            fffqw [@c](/n/{nid1['id']})
            fff
            [@我133](/n/{nid2['id']})
            ffq
            """)
        node, code = await core.node.post(
            au=self.au, md=md, type_=const.NodeTypeEnum.MARKDOWN.value
        )
        self.assertEqual(const.CodeEnum.OK, code)
        await client.search.refresh()
        nodes, total = await client.search.search(au=self.au)
        self.assertEqual(5, len(nodes))
        self.assertEqual(5, total)

        found, total = await client.search.search(au=self.au, query="我")
        self.assertEqual(2, len(found), msg=found)
        self.assertEqual(2, total)

        recommend = await client.search.recommend(
            au=self.au,
            content="I do need a Knowledge Management System. This is a good one to try.",
            exclude_nids=[],
        )
        self.assertEqual(1, len(recommend))

        n, code = await core.node.get(au=self.au, nid=node["id"])
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(2, len(n["toNodeIds"]))

        tmp_text = n["md"]
        n, _, code = await core.node.update_md(au=self.au, nid=node["id"], md=tmp_text + "xxxx")
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(tmp_text + "xxxx", n["md"])

        n, code = await core.node.get(au=self.au, nid=nid1['id'])
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(1, len(n["fromNodeIds"]))

        n, _, code = await core.node.update_md(au=self.au, nid=node["id"], md=n["md"])
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(0, len(n["toNodeIds"]))

        n, code = await core.node.get(au=self.au, nid=nid1['id'])
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(0, len(n["fromNodeIds"]))

    @utils.skip_no_connect
    async def test_add_set(self):
        node, code = await core.node.post(
            au=self.au, md="title\ntext", type_=const.NodeTypeEnum.MARKDOWN.value
        )
        self.assertEqual(0, len(node["toNodeIds"]))
        self.assertEqual(const.CodeEnum.OK, code)

        res = await db_ops.node_add_to_set(node["id"], "toNodeIds", ObjectId())
        self.assertEqual(1, res.modified_count)
        node, code = await core.node.get(au=self.au, nid=node["id"])
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(1, len(node["toNodeIds"]))

    @utils.skip_no_connect
    async def test_to_trash(self):
        n1, code = await core.node.post(
            au=self.au, md="title\ntext", type_=const.NodeTypeEnum.MARKDOWN.value
        )
        self.assertEqual(const.CodeEnum.OK, code)
        _, code = await core.node.post(
            au=self.au, md="title2\ntext", type_=const.NodeTypeEnum.MARKDOWN.value
        )
        self.assertEqual(const.CodeEnum.OK, code)

        code = await core.node.to_trash(self.au, n1["id"])
        self.assertEqual(const.CodeEnum.OK, code)

        ns, total = await core.node.get_nodes_in_trash(au=self.au, page=0, limit=10)
        self.assertEqual(1, len(ns))
        self.assertEqual(1, total)
        self.assertEqual(n1["id"], ns[0]["id"])

        await client.search.refresh()
        ns, total = await client.search.search(self.au)
        self.assertEqual(3, len(ns))
        self.assertEqual(3, total)

        code = await core.node.restore_from_trash(self.au, n1["id"])
        self.assertEqual(const.CodeEnum.OK, code)

        await client.search.refresh()
        nodes, total = await client.search.search(self.au)
        self.assertEqual(4, len(nodes))
        self.assertEqual(4, total)

    @utils.skip_no_connect
    @patch("retk.core.node.backup.__remove_md_all_versions_from_cos")
    @patch("retk.core.node.backup.__remove_md_from_cos")
    @patch("retk.core.node.backup.__get_md_from_cos")
    @patch("retk.core.node.backup.__save_md_to_cos")
    async def test_batch(
            self,
            mock_save_md_to_cos,
            mock_get_md_from_cos,
            mock_remove_md_from_cos,
            mock_remove_md_all_versions_from_cos,
    ):
        mock_save_md_to_cos.return_value = const.CodeEnum.OK
        mock_get_md_from_cos.return_value = ("", const.CodeEnum.OK)
        mock_remove_md_from_cos.return_value = const.CodeEnum.OK
        mock_remove_md_all_versions_from_cos.return_value = const.CodeEnum.OK
        ns = []
        for i in range(10):
            n, code = await core.node.post(
                au=self.au, md=f"title{i}\ntext", type_=const.NodeTypeEnum.MARKDOWN.value
            )
            self.assertEqual(const.CodeEnum.OK, code)
            ns.append(n)

        base_count = 2

        code = await core.node.batch_to_trash(self.au, [n["id"] for n in ns[:4]])
        self.assertEqual(const.CodeEnum.OK, code)

        await client.search.refresh()
        nodes, total = await client.search.search(self.au)
        self.assertEqual(6 + base_count, len(nodes))
        self.assertEqual(6 + base_count, total)

        tns, total = await core.node.get_nodes_in_trash(au=self.au, page=0, limit=10)
        self.assertEqual(4, total)
        self.assertEqual(4, len(tns))

        code = await core.node.restore_batch_from_trash(self.au, [n["id"] for n in tns[:2]])
        self.assertEqual(const.CodeEnum.OK, code)

        await client.search.refresh()
        nodes, total = await client.search.search(self.au)
        self.assertEqual(8 + base_count, len(nodes))
        self.assertEqual(8 + base_count, total)

        code = await core.node.batch_delete(self.au, [n["id"] for n in tns[2:4]])
        self.assertEqual(const.CodeEnum.OK, code)

        tns, total = await core.node.get_nodes_in_trash(au=self.au, page=0, limit=10)
        self.assertEqual(0, total)
        self.assertEqual(0, len(tns))

    @utils.skip_no_connect
    async def test_update_used_space(self):
        u, code = await core.user.get(self.au.u.id)
        base_used_space = u["usedSpace"]
        for delta, value in [
            (100, 100),
            (100, 200),
            (0, 200),
            (-3000, 0),
            (20.1, 20.1),
        ]:
            code = await core.user.update_used_space(self.au.u.id, delta)
            self.assertEqual(const.CodeEnum.OK, code, msg=f"delta: {delta}, value: {value}")
            u, code = await core.user.get(self.au.u.id)
            self.assertEqual(const.CodeEnum.OK, code)
            now = u["usedSpace"] - base_used_space
            if now < 0:
                now = 0
                base_used_space = 0
            self.assertAlmostEqual(value, now, msg=f"delta: {delta}, value: {value}")

    @utils.skip_no_connect
    @patch("retk.core.node.backup.__remove_md_all_versions_from_cos")
    @patch("retk.core.node.backup.__remove_md_from_cos")
    @patch("retk.core.node.backup.__get_md_from_cos")
    @patch("retk.core.node.backup.__save_md_to_cos")
    async def test_md_history(
            self,
            mock_save_md_to_cos,
            mock_get_md_from_cos,
            mock_remove_md_from_cos,
            mock_remove_md_all_versions_from_cos,
    ):
        mock_save_md_to_cos.return_value = const.CodeEnum.OK
        mock_get_md_from_cos.return_value = ("title2\ntext", const.CodeEnum.OK)
        mock_remove_md_from_cos.return_value = const.CodeEnum.OK
        mock_remove_md_all_versions_from_cos.return_value = const.CodeEnum.OK

        bi = config.get_settings().MD_BACKUP_INTERVAL
        config.get_settings().MD_BACKUP_INTERVAL = 0.0001
        n1, code = await core.node.post(
            au=self.au, md="title\ntext", type_=const.NodeTypeEnum.MARKDOWN.value
        )
        self.assertEqual(const.CodeEnum.OK, code)
        time.sleep(0.001)

        _, _, code = await core.node.update_md(
            au=self.au, nid=n1["id"], md="title2\ntext",
        )
        self.assertEqual(const.CodeEnum.OK, code)
        time.sleep(0.001)

        _, _, code = await core.node.update_md(
            au=self.au, nid=n1["id"], md="title3\ntext",
        )
        self.assertEqual(const.CodeEnum.OK, code)

        hist, code = await core.node.get_hist_editions(
            au=self.au,
            nid=n1["id"],
        )
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(2, len(hist))

        hist_md, code = await core.node.get_hist_edition_md(
            au=self.au,
            nid=n1["id"],
            version=hist[1],
        )
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual("title2\ntext", hist_md)

        config.get_settings().MD_BACKUP_INTERVAL = bi

    @utils.skip_no_connect
    async def test_system_notice(self):
        au = deepcopy(self.au)
        au.u.type = const.USER_TYPE.MANAGER.id
        publish_at = datetime.datetime.now()
        doc, code = await core.notice.post_in_manager_delivery(
            au=au,
            title="title",
            content="content",
            recipient_type=const.notice.RecipientTypeEnum.ALL.value,
            batch_type_ids=[],
            publish_at=publish_at,
        )
        self.assertEqual(const.CodeEnum.OK, code)

        docs, total = await core.notice.get_system_notices(0, 10)
        self.assertEqual(1, len(docs))
        self.assertEqual(1, total)
        self.assertEqual(doc["_id"], docs[0]["_id"])
        self.assertEqual("title", docs[0]["title"])
        self.assertEqual("content", docs[0]["snippet"])
        self.assertEqual(publish_at.second, docs[0]["publishAt"].second)
        self.assertFalse(docs[0]["scheduled"])

        await tasks.notice.async_deliver_unscheduled_system_notices()
        time.sleep(0.01)
        docs, total = await core.notice.get_system_notices(0, 10)
        self.assertTrue(docs[0]["scheduled"])

    @utils.skip_no_connect
    async def test_notice(self):
        au = deepcopy(self.au)
        doc, code = await core.notice.post_in_manager_delivery(
            au=au,
            title="title",
            content="content",
            recipient_type=const.notice.RecipientTypeEnum.ALL.value,
            batch_type_ids=[],
            publish_at=None,
        )
        self.assertEqual(const.CodeEnum.NOT_PERMITTED, code)

        au.u.type = const.USER_TYPE.MANAGER.id
        doc, code = await core.notice.post_in_manager_delivery(
            au=au,
            title="title",
            content="content",
            recipient_type=const.notice.RecipientTypeEnum.ALL.value,
            batch_type_ids=[],
            publish_at=None,
        )
        self.assertEqual(const.CodeEnum.OK, code)

        await tasks.notice.async_deliver_unscheduled_system_notices()

        res = await client.coll.notice_system.find().to_list(None)
        self.assertEqual(1, len(res))
        self.assertEqual(doc["_id"], res[0]["noticeId"])
        self.assertEqual(au.u.id, res[0]["senderId"])
        self.assertFalse(res[0]["read"])

        d = await client.coll.notice_manager_delivery.find_one({"_id": doc["_id"]})
        self.assertIsNotNone(d)
        self.assertTrue(d["scheduled"])

        n, code = await core.notice.get_user_notices(au)
        self.assertEqual(const.CodeEnum.OK, code)
        sn = n["system"]["notices"]
        self.assertEqual(1, len(sn))
        self.assertEqual(1, n["system"]["total"])
        self.assertEqual(str(doc["_id"]), sn[0]["id"])
        self.assertEqual("title", sn[0]["title"])
        self.assertEqual("content", sn[0]["snippet"])
        self.assertLess(datetime.datetime.strptime(sn[0]["publishAt"], '%Y-%m-%dT%H:%M:%SZ'), datetime.datetime.now())
        self.assertFalse(sn[0]["read"])
        self.assertIsNone(sn[0]["readTime"])

        n, code = await core.notice.get_system_notice(au.u.id, notice_id=str(doc["_id"]))
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual("title", n["title"])
        self.assertEqual("<p>content</p>", n["html"])
        self.assertLess(datetime.datetime.strptime(sn[0]["publishAt"], '%Y-%m-%dT%H:%M:%SZ'), datetime.datetime.now())
        self.assertFalse(sn[0]["read"])
        self.assertIsNone(sn[0]["readTime"])

    @utils.skip_no_connect
    async def test_mark_read(self):
        au = deepcopy(self.au)
        au.u.type = const.USER_TYPE.MANAGER.id
        for i in range(3):
            _, code = await core.notice.post_in_manager_delivery(
                au=au,
                title=f"title{i}",
                content=f"content{i}",
                recipient_type=const.notice.RecipientTypeEnum.ALL.value,
                batch_type_ids=[],
                publish_at=None,
            )
            self.assertEqual(const.CodeEnum.OK, code)

        await tasks.notice.async_deliver_unscheduled_system_notices()

        au.u.type = const.USER_TYPE.NORMAL.id
        n, code = await core.notice.get_user_notices(au)
        self.assertEqual(const.CodeEnum.OK, code)
        sn = n["system"]["notices"]
        self.assertEqual(3, len(sn))
        self.assertEqual(3, n["system"]["total"])
        for s in sn:
            self.assertFalse(s["read"])
            self.assertIsNone(s["readTime"])

        code = await core.notice.mark_system_notice_read(au.u.id, sn[0]["id"])
        self.assertEqual(const.CodeEnum.OK, code)

        n, code = await core.notice.get_user_notices(au)
        self.assertEqual(const.CodeEnum.OK, code)
        _sn = n["system"]["notices"]
        self.assertEqual(3, len(_sn))
        self.assertEqual(3, n["system"]["total"])
        for i in range(3):
            if _sn[i]["id"] == sn[0]["id"]:
                self.assertTrue(_sn[i]["read"])
                self.assertIsNotNone(_sn[i]["readTime"])
                continue
            self.assertFalse(sn[i]["read"])
            self.assertIsNone(sn[i]["readTime"])

        code = await core.notice.mark_all_system_notice_read(au)
        self.assertEqual(const.CodeEnum.OK, code)

        n, code = await core.notice.get_user_notices(au)
        self.assertEqual(const.CodeEnum.OK, code)
        sn = n["system"]["notices"]
        self.assertEqual(3, len(sn))
        self.assertEqual(3, n["system"]["total"])
        for s in sn:
            self.assertTrue(s["read"])
            self.assertIsNotNone(s["readTime"])



import multiprocessing
import shutil
import socket
import time
import unittest
import urllib.request
from pathlib import Path

import retk
from retk import config, const


class TestRun(unittest.TestCase):
    path: Path

    @classmethod
    def setUpClass(cls) -> None:
        config.get_settings.cache_clear()
        cls.path = p = Path(__file__).parent / "tmp"
        p.mkdir(exist_ok=True)

    @classmethod
    def tearDownClass(cls) -> None:
        shutil.rmtree(str(cls.path), ignore_errors=True)
        config.get_settings.cache_clear()

    def test_run(self, password=None):
        port = 8001

        p = multiprocessing.Process(target=retk.run, kwargs={
            "path": self.path, "port": port, "language": "zh", "headless": True,
            "debug": True, "password": password,
        })
        p.start()
        # p.join()
        while True:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(2)
            result = sock.connect_ex(('127.0.0.1', port))
            sock.close()
            if result == 0:
                break
            time.sleep(0.1)

        for url in [
            "",
            "/r",
            "/r/login",
            "/r/sauth",
            "/r/settings",
            "/r/user",
            "/r/import",
            "/r/n/123",
        ]:
            resp = urllib.request.urlopen(f"http://127.0.0.1:{port}{url}")
            self.assertEqual(200, resp.status, msg=f"failed to get {url}")
            self.assertEqual("text/html; charset=utf-8", resp.headers["content-type"])
            html = resp.read().decode()
            self.assertIn("Rethink", html)
            self.assertIn('<div id="app">', html)

        p.kill()
        p.join()
        self.assertTrue(self.path.exists())
        self.assertTrue(self.path.is_dir())
        self.assertTrue((self.path / ".data").exists())
        self.assertEqual(2, len(list((self.path / const.settings.DOT_DATA / "md").glob("*.md"))))

    def test_run_with_pw(self):
        return self.test_run(password="12345678")

    # def test_plugin(self):
    #
    #     class TestPlugin(retk.Plugin):
    #         name = "TestPlugin"
    #         version = "0.1.0"
    #         description = "A demo test plugin."
    #         author = "morvanzhou"
    #         template = "<h1>{h}</h1>\n<p>{p}</p>"
    #         schedule_timing = retk.schedule.every_minute_at(second=0)
    #
    #         def __init__(self):
    #             super().__init__()
    #             self.count = 0
    #
    #         def on_schedule(self) -> None:
    #             print(self.count, time.time())
    #             self.count += 1
    #
    #     plugin = TestPlugin()
    #     retk.add_plugin(plugin)
    #     retk.run()



import bson

from .collection import Collection
from .command_cursor import CommandCursor
from .common import support_alert, ok_name, MetaStorageObject
from .errors import MongitaNotImplementedError, InvalidName


class Database():
    UNIMPLEMENTED = ['aggregate', 'codec_options', 'command', 'create_collection', 'dereference', 'get_collection',
                     'next', 'profiling_info', 'profiling_level', 'read_concern', 'read_preference',
                     'set_profiling_level', 'validate_collection', 'watch', 'with_options', 'write_concern']
    DEPRECATED = ['add_son_manipulator', 'add_user', 'authenticate', 'collection_names', 'current_op', 'error', 'eval',
                  'incoming_copying_manipulators', 'incoming_manipulators', 'last_status', 'logout',
                  'outgoing_copying_manipulators', 'outgoing_manipulators', 'previous_error', 'remove_user',
                  'reset_error_history', 'system_js', 'SystemJS']

    def __init__(self, db_name, client):
        self.name = db_name
        self.client = client
        self._engine = client.engine
        self._base_location = db_name
        self._cache = {}

    def __repr__(self):
        return "Database(%s, %r)" % (repr(self.client), self.name)

    def __getattr__(self, attr):
        if attr in self.UNIMPLEMENTED:
            raise MongitaNotImplementedError.create("Database", attr)
        if attr in self.DEPRECATED:
            raise MongitaNotImplementedError.create_depr("Database", attr)
        if attr == '_Collection__create':
            return self.__create
        return self[attr]

    def __getitem__(self, collection_name):
        try:
            return self._cache[collection_name]
        except KeyError:
            if not ok_name(collection_name):
                raise InvalidName("Collection cannot be named %r." % collection_name)
            coll = Collection(collection_name, self)
            self._cache[collection_name] = coll
            return coll

    def __create(self, coll_name):
        """
        Mongodb does not create anything until first insert. So when we insert
        something, this will create a small metadata file to basically just store
        our collection_names
        """
        metadata = self._engine.get_metadata(self._base_location)
        if metadata:
            if coll_name not in metadata['collection_names']:
                metadata['collection_names'].append(coll_name)
                assert self._engine.put_metadata(self._base_location, metadata)
            return
        self._engine.create_path(self._base_location)
        metadata = MetaStorageObject({
            'options': {},
            'collection_names': [coll_name],
            'uuid': str(bson.ObjectId()),
        })
        assert self._engine.put_metadata(self._base_location, metadata)
        self.client.__create(self.name)

    @support_alert
    def list_collection_names(self):
        """
        List every collection name.

        :rtype: list[str]
        """
        metadata = self._engine.get_metadata(self._base_location)
        if metadata:
            return metadata['collection_names']
        return []

    @support_alert
    def list_collections(self):
        """
        Returns a cursor to iterate over all collections.

        :rtype: CommandCursor
        """

        def cursor():
            for coll_name in self.list_collection_names():
                if coll_name not in self._cache:
                    self._cache[coll_name] = Collection(coll_name, self)
                yield self._cache[coll_name]

        return CommandCursor(cursor)

    @support_alert
    def drop_collection(self, name_or_collection):
        """
        Drop a collection.

        :param name_or_collection str|collection.Collection:
        :rtype: None
        """
        if isinstance(name_or_collection, Collection):
            collection = name_or_collection.name
        else:
            collection = name_or_collection
        self._engine.delete_dir(f'{self.name}.{collection}')
        metadata = self._engine.get_metadata(self._base_location)
        if metadata and collection in metadata['collection_names']:
            metadata['collection_names'].remove(collection)
            assert self._engine.put_metadata(self._base_location, metadata)
        try:
            self._cache[collection]._existence_verified = False
            del self._cache[collection]
        except KeyError:
            pass



"""Google SSO Login Helper
"""

import httpx

from .base import DiscoveryDocument, OpenID, SSOBase, SSOLoginError


class QQSSO(SSOBase):
    """Class providing login via Google OAuth"""

    discovery_url = "https://accounts.google.com/.well-known/openid-configuration"
    provider = "tencent"
    scope = ["openid", "email", "profile"]

    @classmethod
    async def openid_from_response(cls, response: dict) -> OpenID:
        """Return OpenID from user information provided by Google"""
        if response.get("email_verified"):
            return OpenID(
                email=response.get("email", ""),
                provider=cls.provider,
                id=response.get("sub"),
                first_name=response.get("given_name"),
                last_name=response.get("family_name"),
                display_name=response.get("name"),
                picture=response.get("picture"),
            )
        raise SSOLoginError(401, f"User {response.get('email')} is not verified with Google")

    async def get_discovery_document(self) -> DiscoveryDocument:
        """Get document containing handy urls"""
        async with httpx.AsyncClient() as session:
            response = await session.get(self.discovery_url)
            content = response.json()
            return content



import asyncio
import time
import unittest
from datetime import datetime, timedelta

from bson.tz_util import utc

from retk.const.settings import MAX_SCHEDULE_JOB_INFO_LEN
from retk.core import scheduler
from retk.logger import logger


def print_test(txt: str):
    text = f"test and print '{txt}'"
    print(text)
    return text


def fake_test(inp):
    return inp


def async_test(txt: str):
    async def async_test_inner(txt):
        text = f"async_test and print '{txt}'"
        print(text)
        return text

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    res = loop.run_until_complete(async_test_inner(txt))
    loop.close()
    return res


class TestCoreAsyncTask(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        scheduler.start()

    def setUp(self) -> None:
        scheduler.clear_jobs()
        self.log_level = logger.level
        logger.setLevel("DEBUG")

    def tearDown(self) -> None:
        logger.setLevel(self.log_level)

    @classmethod
    def tearDownClass(cls) -> None:
        scheduler.stop()

    def test_put_task(self):
        scheduler.run_once_after(
            job_id="test1",
            second=0,
            func=print_test,
            args=("1",),
        )
        scheduler.run_once_after(
            job_id="test2",
            func=print_test,
            second=0.001,
            args=("2",),
        )
        scheduler.run_once_at(
            job_id="test3",
            func=print_test,
            time=datetime.now(tz=utc),
            args=("3",),
        )
        scheduler.run_once_at(
            job_id="test4",
            func=print_test,
            time=datetime.now(tz=utc) + timedelta(seconds=0.003),
            args=("4",),
        )

        time.sleep(0.005)
        now = datetime.now(tz=utc)
        scheduler.run_every_at(
            job_id="test5",
            func=print_test,
            second=(now.second + 1) % 60,
            args=("5",),
        )

        self.assertEqual(5, len(scheduler.get_jobs()))
        for i, job in enumerate(scheduler.get_jobs()[::-1]):
            if i == 0:
                self.assertIsNone(job.finished_at, msg=job)
                continue
            fail = True
            for _ in range(10):
                if job.finished_at is None:
                    time.sleep(0.01)
                    continue
                self.assertIsInstance(job.finished_at, datetime)
                self.assertTrue(job.finished_return.startswith("test and print"))
                fail = False
            self.assertFalse(fail)

        time.sleep(1.1)

        for _, job in enumerate(scheduler.get_jobs()):
            self.assertIsInstance(job.finished_at, datetime)
            self.assertTrue(job.finished_return.startswith("test and print"))

        j = scheduler.get_job("test5")
        self.assertIsInstance(j, scheduler.schedule.JobInfo)
        self.assertEqual("test and print '5'", j.finished_return)

        with self.assertRaises(KeyError) as e:
            scheduler.run_once_now(
                job_id="test5",
                func=fake_test,
                args=(6,),
            )
            self.assertEqual("Job 'test5' already exists", str(e.exception))

    def test_async_task(self):
        scheduler.run_once_now(
            job_id="test1",
            func=async_test,
            args=("1",),
        )
        scheduler.run_once_now(
            job_id="test2",
            func=async_test,
            args=("2",),
        )
        self.assertEqual(2, len(scheduler.get_jobs()))
        for job in scheduler.get_jobs():
            fail = True
            for _ in range(10):
                if job.finished_at is None:
                    time.sleep(0.01)
                    continue
                self.assertIsInstance(job.finished_at, datetime)
                self.assertTrue(job.finished_return.startswith("async_test and print"))
                fail = False
            self.assertFalse(fail)

    def test_max_len(self):
        for i in range(MAX_SCHEDULE_JOB_INFO_LEN + 2):
            scheduler.run_once_now(
                job_id=f"test{i}",
                func=fake_test,
                args=(i,),
            )
        self.assertEqual(MAX_SCHEDULE_JOB_INFO_LEN, len(scheduler.get_jobs()))
        count = 2
        for j in scheduler.get_jobs():
            fail = True
            for _ in range(10):
                if j.finished_at is None:
                    time.sleep(0.01)
                    continue
                self.assertIsInstance(j.created_at, datetime)
                self.assertIsInstance(j.executed_at, datetime)
                self.assertGreaterEqual(j.executing_time(), timedelta(seconds=0))
                self.assertGreaterEqual(j.finished_at, j.executed_at)
                self.assertGreaterEqual(j.executed_at, j.created_at)
                self.assertEqual(count, j.finished_return, msg=str(j))
                fail = False
                break
            self.assertFalse(fail)
            count += 1



import hashlib
import io
from dataclasses import dataclass
from typing import BinaryIO

from PIL import Image, UnidentifiedImageError
from bson import ObjectId

try:
    from qcloud_cos import CosConfig, CosServiceError, CosS3Client
except ImportError:
    pass

from retk.config import get_settings, is_local_db
from retk.const.app import FileTypesEnum
from retk.const.settings import IMG_RESIZE_THRESHOLD, DOT_DATA
from retk.core.user import update_used_space
from retk.logger import logger
from retk.models.client import client
from retk.models.tps import UserFile


@dataclass
class File:
    data: BinaryIO
    filename: str

    def __post_init__(self):
        self.data.seek(0)
        md5_hash = hashlib.md5()
        for chunk in iter(lambda: self.data.read(4096), b""):
            md5_hash.update(chunk)
        self.data.seek(0)
        self.hash = md5_hash.hexdigest()
        sep = self.filename.rsplit(".", 1)
        self.ext = f".{sep[-1]}" if len(sep) > 1 else ""
        self.type = FileTypesEnum.get_type(self.ext)
        self.hashed_filename = f"{self.hash}{self.ext}"
        self._reset_size()

    def image_resize(self, resize_threshold: int):
        if self.type != FileTypesEnum.IMAGE:
            return
        if self.size > resize_threshold:
            # reduce image size
            out = io.BytesIO()
            image = Image.open(self.data)
            image.save(out, format=self.ext.upper()[1:], quality=50, optimize=True)
            self.data = out
        self._reset_size()

    def _reset_size(self):
        self.data.seek(0, io.SEEK_END)
        self.size = self.data.tell()
        self.data.seek(0)

    def is_unknown_type(self):
        return self.type == FileTypesEnum.UNKNOWN


async def add_to_db(
        uid: str,
        file: File,
):
    doc: UserFile = {
        "_id": ObjectId(),
        "uid": uid,
        "fid": file.hash,
        "filename": file.filename,
        "size": file.size,
    }
    await client.coll.user_file.insert_one(doc)
    await update_used_space(uid=uid, delta=file.size)


class Saver:
    resize_threshold = IMG_RESIZE_THRESHOLD

    async def save(self, uid: str, file: File):
        if is_local_db():
            url = await self.save_local(uid=uid, file=file)
        else:
            url = await self.save_remote(uid=uid, file=file)
        return url

    async def save_local(self, uid: str, file: File) -> str:
        path = get_settings().RETHINK_LOCAL_STORAGE_PATH / DOT_DATA / "files" / file.hashed_filename
        path.parent.mkdir(parents=True, exist_ok=True)

        if path.exists():
            # skip the same image
            return f"/files/{file.hashed_filename}"

        try:
            if file.type == FileTypesEnum.IMAGE:
                file.image_resize(resize_threshold=self.resize_threshold)
                Image.open(file.data).save(path)
            else:
                with open(path, "wb") as f:
                    f.write(file.data.read())
        except (
                FileNotFoundError,
                OSError,
                UnidentifiedImageError,
                ValueError,
                TypeError,
        ) as e:
            logger.error(f"failed to save file: {file.filename}. {e}")
            return ""

        await add_to_db(uid=uid, file=file)
        return f"/files/{file.hashed_filename}"

    async def save_remote(self, uid: str, file: File):
        # to cos
        token = None

        settings = get_settings()
        secret_id = settings.COS_SECRET_ID
        secret_key = settings.COS_SECRET_KEY
        region = settings.COS_REGION
        domain = None
        cos_conf = CosConfig(
            Region=region,
            SecretId=secret_id,
            SecretKey=secret_key,
            Token=token,
            Domain=domain,
            Scheme='https',
        )
        cos_client = CosS3Client(cos_conf)

        key = f"userData/{uid}/{file.hashed_filename}"

        url = f"https://{settings.COS_BUCKET_NAME}.cos.{settings.COS_REGION}.myqcloud.com/{key}"

        doc = await client.coll.user_file.find_one({"uid": uid, "fid": file.hashed_filename})
        if doc:
            return url

        try:
            _ = cos_client.head_object(
                Bucket=settings.COS_BUCKET_NAME,
                Key=key
            )
            return url
        except CosServiceError as e:
            if e.get_status_code() != 404:
                return url

        if file.type == FileTypesEnum.IMAGE:
            file.image_resize(resize_threshold=self.resize_threshold)

        # can raise error
        try:
            _ = cos_client.put_object(
                Bucket=settings.COS_BUCKET_NAME,
                Body=file.data,
                Key=key,
                StorageClass='STANDARD',  # 'STANDARD'|'STANDARD_IA'|'ARCHIVE',
                EnableMD5=False,
                # ContentType=content_type,
            )
        except CosServiceError as e:
            logger.error(f"failed to save file to cos: {e}")
            return ""
        await add_to_db(uid=uid, file=file)
        return url


saver = Saver()



from retk import core
from retk.controllers import schemas
from retk.controllers.utils import maybe_raise_json_exception
from retk.models.tps import AuthedUser
from retk.utils import datetime2str


async def get_system_notice(
        au: AuthedUser,
        notice_id: str,
) -> schemas.notice.SystemNoticeResponse:
    code = await core.notice.mark_system_notice_read(
        uid=au.u.id,
        notice_id=notice_id,
    )
    maybe_raise_json_exception(au=au, code=code)

    n, code = await core.notice.get_system_notice(
        uid=au.u.id,
        notice_id=notice_id,
    )
    maybe_raise_json_exception(au=au, code=code)

    return schemas.notice.SystemNoticeResponse(
        requestId=au.request_id,
        notice=schemas.notice.SystemNoticeResponse.Notice(
            id=str(n["_id"]),
            title=n["title"],
            html=n["html"],
            publishAt=datetime2str(n["publishAt"]),
        ),
    )



from fastapi import APIRouter

from retk.controllers import schemas, recent
from retk.routes import utils

router = APIRouter(
    prefix="/api/recent",
    tags=["search"],
    responses={404: {"description": "Not found"}},
)


@router.post(
    path="/at",
    status_code=201,
    response_model=schemas.RequestIdResponse,
)
@utils.measure_time_spend
async def add_recent_at_node(
        au: utils.ANNOTATED_AUTHED_USER,
        req: schemas.recent.AtNodeRequest,
) -> schemas.RequestIdResponse:
    return await recent.add_recent_at_node(
        au=au,
        req=req,
    )


@router.get(
    path="/searched",
    status_code=200,
    response_model=schemas.recent.GetRecentSearchResponse,
)
@utils.measure_time_spend
async def get_recent_searched(
        au: utils.ANNOTATED_AUTHED_USER,
) -> schemas.recent.GetRecentSearchResponse:
    return await recent.get_recent_searched(
        au=au,
    )



from typing import Optional

from fastapi import APIRouter
from fastapi.responses import StreamingResponse

from retk.controllers import account
from retk.routes import utils

router = APIRouter(
    prefix="/api/captcha",
    tags=["captcha"],
    responses={404: {"description": "Not found"}},
)


@router.get(
    path="/img",
    status_code=200,
)
@utils.measure_time_spend
async def get_captcha_img(
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> StreamingResponse:
    return account.get_captcha_img()



"""Gitlab SSO Oauth Helper class"""

from .base import DiscoveryDocument, OpenID, SSOBase


class GitlabSSO(SSOBase):
    """Class providing login via Gitlab SSO"""

    provider = "gitlab"
    scope = ["read_user", "openid", "profile"]
    additional_headers = {"accept": "application/json"}

    async def get_discovery_document(self) -> DiscoveryDocument:
        return {
            "authorization_endpoint": "https://gitlab.com/oauth/authorize",
            "token_endpoint": "https://gitlab.com/oauth/token",
            "userinfo_endpoint": "https://gitlab.com/api/v4/user",
        }

    @classmethod
    async def openid_from_response(cls, response: dict) -> OpenID:
        return OpenID(
            email=response["email"],
            provider=cls.provider,
            id=response["id"],
            display_name=response["username"],
            picture=response["avatar_url"],
        )



from dataclasses import dataclass
from enum import IntEnum, unique
from typing import Dict

from .languages import LanguageEnum


@unique
class CodeEnum(IntEnum):
    OK = 0
    ACCOUNT_OR_PASSWORD_ERROR = 1
    INVALID_AUTH = 2
    EXPIRED_AUTH = 3
    USER_EXIST = 4
    NODE_EXIST = 5
    NODE_NOT_EXIST = 6
    OPERATION_FAILED = 7
    EMAIL_OCCUPIED = 8
    PLUGIN_NOT_FOUND = 9
    COS_ERROR = 10
    OLD_PASSWORD_ERROR = 11
    ONE_USER_MODE = 12
    INVALID_PASSWORD = 13
    CAPTCHA_ERROR = 14
    CAPTCHA_EXPIRED = 15
    NOTE_EXCEED_MAX_LENGTH = 16
    INVALID_NODE_DISPLAY_METHOD = 17
    TOO_MANY_FILES = 18
    TOO_LARGE_FILE = 19
    INVALID_FILE_TYPE = 20
    FILE_OPEN_ERROR = 21
    INVALID_SETTING = 22
    IMPORT_PROCESS_NOT_FINISHED = 23
    UPLOAD_TASK_TIMEOUT = 24
    USER_SPACE_NOT_ENOUGH = 25
    INVALID_EMAIL = 26
    URL_TOO_LONG = 27
    OAUTH_PROVIDER_NOT_FOUND = 28
    TASK_NOT_FOUND = 29
    ACCOUNT_EXIST_TRY_FORGET_PASSWORD = 30
    USER_DISABLED = 31
    NOT_PERMITTED = 32
    EXPIRED_OR_NO_ACCESS_TOKEN = 33
    USER_NOT_EXIST = 34
    INVALID_PARAMS = 35
    INVALID_SCHEDULE_JOB_ID = 36
    NOTICE_NOT_FOUND = 37


@dataclass
class CodeMessage:
    zh: str
    en: str


CODE_MESSAGES: Dict[CodeEnum, CodeMessage] = {
    CodeEnum.OK: CodeMessage(zh="成功", en="OK"),
    CodeEnum.ACCOUNT_OR_PASSWORD_ERROR: CodeMessage(zh="账号不存在或者密码错误", en="No such user or password error"),
    CodeEnum.INVALID_AUTH: CodeMessage(zh="无效的认证信息", en="Invalid authentication information"),
    CodeEnum.EXPIRED_AUTH: CodeMessage(zh="认证信息已过期", en="Authentication information has expired"),
    CodeEnum.USER_EXIST: CodeMessage(zh="用户已存在", en="User already exists"),
    CodeEnum.NODE_EXIST: CodeMessage(zh="节点已存在", en="Node already exists"),
    CodeEnum.NODE_NOT_EXIST: CodeMessage(zh="节点不存在", en="Node does not exist"),
    CodeEnum.OPERATION_FAILED: CodeMessage(zh="操作失败", en="Operation failed"),
    CodeEnum.EMAIL_OCCUPIED: CodeMessage(zh="邮箱已被占用", en="Email is occupied"),
    CodeEnum.PLUGIN_NOT_FOUND: CodeMessage(zh="插件未找到", en="Plugin not found"),
    CodeEnum.COS_ERROR: CodeMessage(zh="COS 错误", en="COS error"),
    CodeEnum.OLD_PASSWORD_ERROR: CodeMessage(zh="旧密码错误", en="Old password error"),
    CodeEnum.ONE_USER_MODE: CodeMessage(zh="单用户模式，不支持注册",
                                        en="Single user mode, registration is not supported"),
    CodeEnum.INVALID_PASSWORD: CodeMessage(zh="密码格式错误", en="Password format error"),
    CodeEnum.CAPTCHA_ERROR: CodeMessage(zh="验证码输入错误", en="Captcha not match"),
    CodeEnum.CAPTCHA_EXPIRED: CodeMessage(zh="验证码已过期", en="Captcha expired"),
    CodeEnum.NOTE_EXCEED_MAX_LENGTH: CodeMessage(zh="内容超过最大长度", en="Content exceed max length"),
    CodeEnum.INVALID_NODE_DISPLAY_METHOD: CodeMessage(zh="无效的展示方式", en="Invalid display method"),
    CodeEnum.TOO_MANY_FILES: CodeMessage(zh="文件数量过多", en="Too many files"),
    CodeEnum.TOO_LARGE_FILE: CodeMessage(zh="文件过大", en="Too large file"),
    CodeEnum.INVALID_FILE_TYPE: CodeMessage(zh="无效的文件类型", en="Invalid file type"),
    CodeEnum.FILE_OPEN_ERROR: CodeMessage(zh="文件打开失败", en="File open error"),
    CodeEnum.INVALID_SETTING: CodeMessage(zh="无效的设置", en="Invalid setting"),
    CodeEnum.IMPORT_PROCESS_NOT_FINISHED: CodeMessage(
        zh="正在完成上一批数据导入，请稍后再试",
        en="Last import process not finished, please try again later"),
    CodeEnum.UPLOAD_TASK_TIMEOUT: CodeMessage(zh="文件上传任务超时", en="Upload task timeout"),
    CodeEnum.USER_SPACE_NOT_ENOUGH: CodeMessage(zh="用户空间不足", en="User space not enough"),
    CodeEnum.INVALID_EMAIL: CodeMessage(zh="邮箱格式错误", en="Email format error"),
    CodeEnum.URL_TOO_LONG: CodeMessage(zh="URL字符太长", en="URL too long"),
    CodeEnum.OAUTH_PROVIDER_NOT_FOUND: CodeMessage(zh="未找到 OAuth 提供商", en="OAuth provider not found"),
    CodeEnum.TASK_NOT_FOUND: CodeMessage(zh="任务未找到", en="Task not found"),
    CodeEnum.ACCOUNT_EXIST_TRY_FORGET_PASSWORD: CodeMessage(
        zh="账户已存在，请尝试通过忘记密码找回",
        en="Account exists, try forget password to recover",
    ),
    CodeEnum.USER_DISABLED: CodeMessage(
        zh="因违反平台规则，此账户已被禁用",
        en="This account has been disabled due to violation of platform rules"
    ),
    CodeEnum.NOT_PERMITTED: CodeMessage(zh="无权限", en="Not permitted"),
    CodeEnum.EXPIRED_OR_NO_ACCESS_TOKEN: CodeMessage(zh="访问令牌已过期或失效",
                                                     en="Access token has expired or invalid"),
    CodeEnum.USER_NOT_EXIST: CodeMessage(zh="用户不存在", en="User does not exist"),
    CodeEnum.INVALID_PARAMS: CodeMessage(zh="无效参数", en="Invalid parameter"),
    CodeEnum.INVALID_SCHEDULE_JOB_ID: CodeMessage(zh="无效的任务 ID", en="Invalid schedule job ID"),
    CodeEnum.NOTICE_NOT_FOUND: CodeMessage(zh="通知未找到", en="Notice not found"),
}

CODE2STATUS_CODE: Dict[CodeEnum, int] = {
    CodeEnum.OK: 200,
    CodeEnum.ACCOUNT_OR_PASSWORD_ERROR: 401,
    CodeEnum.INVALID_AUTH: 401,
    CodeEnum.EXPIRED_AUTH: 401,
    CodeEnum.USER_EXIST: 422,
    CodeEnum.NODE_EXIST: 422,
    CodeEnum.NODE_NOT_EXIST: 404,
    CodeEnum.OPERATION_FAILED: 500,
    CodeEnum.EMAIL_OCCUPIED: 422,
    CodeEnum.PLUGIN_NOT_FOUND: 404,
    CodeEnum.COS_ERROR: 500,
    CodeEnum.OLD_PASSWORD_ERROR: 400,
    CodeEnum.ONE_USER_MODE: 403,
    CodeEnum.INVALID_PASSWORD: 400,
    CodeEnum.CAPTCHA_ERROR: 400,
    CodeEnum.CAPTCHA_EXPIRED: 400,
    CodeEnum.NOTE_EXCEED_MAX_LENGTH: 400,
    CodeEnum.INVALID_NODE_DISPLAY_METHOD: 400,
    CodeEnum.TOO_MANY_FILES: 400,
    CodeEnum.TOO_LARGE_FILE: 400,
    CodeEnum.INVALID_FILE_TYPE: 400,
    CodeEnum.FILE_OPEN_ERROR: 400,
    CodeEnum.INVALID_SETTING: 400,
    CodeEnum.IMPORT_PROCESS_NOT_FINISHED: 403,
    CodeEnum.UPLOAD_TASK_TIMEOUT: 408,
    CodeEnum.USER_SPACE_NOT_ENOUGH: 403,
    CodeEnum.INVALID_EMAIL: 400,
    CodeEnum.URL_TOO_LONG: 406,
    CodeEnum.OAUTH_PROVIDER_NOT_FOUND: 404,
    CodeEnum.TASK_NOT_FOUND: 404,
    CodeEnum.ACCOUNT_EXIST_TRY_FORGET_PASSWORD: 422,
    CodeEnum.USER_DISABLED: 403,
    CodeEnum.NOT_PERMITTED: 403,
    CodeEnum.EXPIRED_OR_NO_ACCESS_TOKEN: 200,
    CodeEnum.USER_NOT_EXIST: 404,
    CodeEnum.INVALID_PARAMS: 400,
    CodeEnum.INVALID_SCHEDULE_JOB_ID: 400,
    CodeEnum.NOTICE_NOT_FOUND: 404,
}


def get_msg_by_code(code: CodeEnum, language: str = LanguageEnum.EN.value):
    msg = CODE_MESSAGES[code]
    if language == LanguageEnum.ZH.value:
        return msg.zh
    elif language == LanguageEnum.EN.value:
        return msg.en
    else:
        raise ValueError(f"Invalid language: {language}")



from retk import config
from retk.logger import logger
from retk.plugins.base import add_plugin, remove_plugin
from retk.plugins.official_plugins.favorites.main import Favorites
from retk.plugins.official_plugins.summary.main import DailySummary

_official_plugins = [
    DailySummary,
    Favorites,
]

_registered_official_plugins = []


def register_official_plugins():
    if not config.get_settings().PLUGINS:
        return

    for _p in _official_plugins:
        _p_instance = _p()
        add_plugin(_p_instance)
        _registered_official_plugins.append(_p_instance)
        logger.debug(f"added plugin '{_p.name}' (id={_p.id})")


def unregister_official_plugins():
    if not config.get_settings().PLUGINS:
        return

    for _p in _registered_official_plugins:
        remove_plugin(_p)
        logger.debug(f"removed plugin '{_p.name}' (id={_p.id})")



import collections
import itertools
import os
import pathlib
import shutil
import threading
from sys import intern as itrn

import bson

from .engine_common import Engine
from ..common import MetaStorageObject, secure_filename

DISK_ENGINE_INCUMBENTS = {}


class DiskEngine(Engine):
    def __init__(self, base_storage_path):
        if not os.path.exists(base_storage_path):
            os.mkdir(base_storage_path)
        self.base_storage_path = base_storage_path
        self._cache = collections.defaultdict(dict)
        self._collection_fhs = {}
        self._metadata = {}
        self._file_attrs = collections.defaultdict(dict)
        self.replaced = False
        self.lock = threading.RLock()

    @staticmethod
    def create(base_storage_path):
        if base_storage_path in DISK_ENGINE_INCUMBENTS:
            DISK_ENGINE_INCUMBENTS[base_storage_path].close()
            return DISK_ENGINE_INCUMBENTS[base_storage_path]
        de = DiskEngine(base_storage_path)
        DISK_ENGINE_INCUMBENTS[base_storage_path] = de
        return de

    def _get_full_path(self, collection, filename=''):
        return os.path.join(*list(filter(None, (self.base_storage_path,
                                                secure_filename(collection),
                                                filename))))

    def _get_coll_fh(self, collection):
        try:
            return self._collection_fhs[collection]
        except KeyError:
            pass
        data_path = self._get_full_path(collection, '$.data')
        if not os.path.exists(data_path):
            # self.create_path(collection)
            pathlib.Path(data_path).touch()
        fh = open(data_path, 'rb+')
        self._collection_fhs[itrn(collection)] = fh
        return fh

    def _get_file_attrs(self, collection):
        if collection in self._file_attrs:
            return self._file_attrs[collection]['loc_idx']

        file_attrs_path = self._get_full_path(collection, '$.file_attrs')
        try:
            with open(file_attrs_path, 'rb') as f:
                self._file_attrs[itrn(collection)] = bson.decode(f.read())
        except FileNotFoundError:
            self._file_attrs[itrn(collection)] = {
                'loc_idx': {},
                'spare_bytes': 0,
                'total_bytes': 0
            }
        return self._file_attrs[collection]['loc_idx']

    def _set_file_attrs(self, collection, doc_id, pos):
        if pos is None:
            self._file_attrs[itrn(collection)]['loc_idx'].pop(doc_id, None)
        else:
            self._file_attrs[itrn(collection)]['loc_idx'][itrn(doc_id)] = pos

    def doc_exists(self, collection, doc_id):
        if str(doc_id) in self._get_file_attrs(collection):
            return True
        return False

    def get_doc(self, collection, doc_id):
        doc_id = str(doc_id)
        try:
            return self._cache[itrn(collection)][itrn(doc_id)]
        except KeyError:
            pass

        pos = self._get_file_attrs(collection)[str(doc_id)]
        fh = self._get_coll_fh(collection)
        fh.seek(pos)
        doc_len_bytes = fh.read(4)
        doc_len = int.from_bytes(doc_len_bytes, 'little', signed=True)
        assert doc_len
        doc = bson.decode(doc_len_bytes + fh.read(doc_len - 4))
        self._cache[itrn(collection)][itrn(doc_id)] = doc
        return doc

    def put_doc(self, collection, doc, no_overwrite=False):
        doc_id = str(doc['_id'])
        if no_overwrite and self.doc_exists(collection, doc_id):
            return False
        self._cache[itrn(collection)][itrn(doc_id)] = doc

        encoded_doc = bson.encode(doc)
        fh = self._get_coll_fh(collection)
        pos = self._get_file_attrs(collection).get(str(doc_id))
        if pos is not None:
            fh.seek(pos)
            doc_len_bytes = fh.read(4)
            doc_len = int.from_bytes(doc_len_bytes, 'little', signed=True)
            assert doc_len
            spare_bytes = doc_len - len(encoded_doc)
            if spare_bytes >= 0:
                fh.seek(pos)
                fh.write(encoded_doc + b'\x00' * spare_bytes)
                fh.flush()
                self._file_attrs[collection]['spare_bytes'] += spare_bytes
                self._file_attrs[collection]['total_bytes'] -= spare_bytes
                return True
        fh.seek(0, 2)
        pos = fh.tell()
        fh.write(encoded_doc)
        fh.flush()
        self._set_file_attrs(collection, doc_id, pos)
        self._file_attrs[collection]['total_bytes'] += len(encoded_doc)
        return True

    def delete_doc(self, collection, doc_id):
        doc_id = str(doc_id)
        pos = self._get_file_attrs(collection)[doc_id]
        fh = self._get_coll_fh(collection)
        fh.seek(pos)
        doc_len_bytes = fh.read(4)
        doc_len = int.from_bytes(doc_len_bytes, 'little', signed=True)
        assert doc_len
        fh.seek(pos)
        fh.write(b'\x00' * doc_len)
        fh.flush()
        self._set_file_attrs(collection, doc_id, None)
        self._file_attrs[collection]['total_bytes'] -= doc_len
        self._file_attrs[collection]['spare_bytes'] += doc_len
        self._cache[collection].pop(doc_id, None)
        return True

    def get_metadata(self, collection):
        try:
            return self._metadata[collection]
        except KeyError:
            pass

        metadata_path = self._get_full_path(collection, '$.metadata')
        try:
            with open(metadata_path, 'rb') as f:
                metadata = MetaStorageObject.from_storage(f.read(), from_bson=True)
        except FileNotFoundError:
            return None
        self._metadata[collection] = metadata
        return metadata

    # TODO disaster recovery rebuilds
    # def _rebuild_metadata(self, coll_path):
    #     fh = self._get_coll_fh(coll_path)
    #     pos = 0
    #     fh.seek(0)
    #     docs = []
    #     while True:
    #         doc_len_bytes = fh.read(4)
    #         if not doc_len_bytes:
    #             break
    #         doc_len = int.from_bytes(doc_len_bytes, 'little', signed=True)
    #         if not doc_len:
    #             continue
    #         doc = bson.decode(doc_len_bytes + fh.read(doc_len - 4))
    #         docs.append((pos, doc))
    #         pos += doc_len

    #     metadata = {}
    #     for pos, doc in docs:
    #         metadata['file_attrs'][doc['_id']] = pos
    #         self._cache[coll_path][doc['_id']] = doc
    #     self._metadata[coll_path] = metadata
    #     return metadata
    def _defrag(self, collection):
        fh = self._get_coll_fh(collection)
        encoded_docs = {}
        _cache_collection = self._cache[itrn(collection)]
        for doc_id in self.list_ids(collection):
            try:
                doc = _cache_collection[itrn(doc_id)]
                encoded_docs[doc_id] = bson.encode(doc)
            except KeyError:
                pos = self._get_file_attrs(collection)[str(doc_id)]
                fh = self._get_coll_fh(collection)
                fh.seek(pos)
                doc_len_bytes = fh.read(4)
                doc_len = int.from_bytes(doc_len_bytes, 'little', signed=True)
                assert doc_len
                encoded_docs[doc_id] = doc_len_bytes + fh.read(doc_len - 4)

        pos = 0
        fh.seek(0)
        for doc_id, encoded_doc in encoded_docs.items():
            fh.write(encoded_doc)
            self._set_file_attrs(collection, doc_id, pos)
            pos += len(encoded_doc)
        fh.truncate()

    def put_metadata(self, collection, metadata):
        self._metadata[itrn(collection)] = metadata
        metadata_path = self._get_full_path(collection, '$.metadata')
        with open(metadata_path, 'wb') as f:
            f.write(metadata.to_storage(as_bson=True))
        file_attrs_path = self._get_full_path(collection, '$.file_attrs')
        if self._file_attrs.get(collection, {}).get('spare_bytes', 0) / \
                (1 + self._file_attrs.get(collection, {}).get('total_bytes', 0)) > 0.5:
            self._defrag(collection)
        with open(file_attrs_path, 'wb') as f:
            f.write(bson.encode(self._file_attrs.get(collection, {'total_bytes': 0,
                                                                  'spare_bytes': 0,
                                                                  'loc_idx': {}})))
        return True

    def delete_dir(self, collection):
        full_path = self._get_full_path(collection)
        if not os.path.isdir(full_path):
            return False
        shutil.rmtree(full_path)
        self._cache.pop(collection, None)
        self._metadata.pop(collection, None)
        self._file_attrs.pop(collection, None)
        if collection in self._collection_fhs:
            self._collection_fhs[collection].close()
            self._collection_fhs.pop(collection, None)
        return True

    def list_ids(self, collection, limit=None):
        keys = self._get_file_attrs(collection).keys()
        if limit is None:
            return list(map(str, keys))
        return list(map(str, itertools.islice(keys, limit)))

    def create_path(self, collection):
        full_loc = self._get_full_path(collection)
        if not os.path.exists(full_loc):
            os.makedirs(full_loc)

    def close(self):
        self._cache = collections.defaultdict(dict)
        self._metadata = {}
        self._file_attrs = {}
        for fh in self._collection_fhs.values():
            fh.close()
        self._collection_fhs = {}



from bson import ObjectId

from retk import const
from retk.models.client import client


async def add_user_behavior(
        uid: str,
        type_: const.UserBehaviorTypeEnum,
        remark: str,
):
    await client.coll.user_behavior.insert_one({
        "_id": ObjectId(),
        "uid": uid,
        "type": type_.value,
        "remark": remark,
    })



from typing import List

from pydantic import BaseModel, NonNegativeInt, Field

from retk.const import settings


class NodeData(BaseModel):
    class LinkedNode(BaseModel):
        id: str
        title: str
        md: str
        snippet: str
        type: NonNegativeInt
        disabled: bool
        inTrash: bool
        createdAt: str
        modifiedAt: str

    id: str
    md: str
    title: str
    snippet: str
    type: NonNegativeInt
    disabled: bool
    createdAt: str
    modifiedAt: str
    fromNodes: List[LinkedNode] = Field(default_factory=list)
    toNodes: List[LinkedNode] = Field(default_factory=list)


class CreateRequest(BaseModel):
    md: str = Field(max_length=settings.MD_MAX_LENGTH)
    type: NonNegativeInt
    fromNid: str = Field(default="", max_length=settings.NID_MAX_LENGTH)


class NodeResponse(BaseModel):
    requestId: str
    node: NodeData


class PatchMdRequest(BaseModel):
    md: str = Field(max_length=settings.MD_MAX_LENGTH)


class NodesSearchResponse(BaseModel):
    class Data(BaseModel):
        class Node(BaseModel):
            id: str
            title: str
            snippet: str
            titleHighlight: str
            bodyHighlights: List[str]
            score: float
            type: int
            createdAt: str
            modifiedAt: str

        nodes: List[Node]
        total: NonNegativeInt

    requestId: str
    data: Data


class BatchNodeIdsRequest(BaseModel):
    nids: List[str] = Field(default_factory=list, min_length=1, max_length=1000)


class HistEditionsResponse(BaseModel):
    requestId: str
    versions: List[str] = Field(default_factory=list)


class HistEditionMdResponse(BaseModel):
    requestId: str
    md: str



class MongitaError(Exception):
    pass


class MongitaNotImplementedError(MongitaError, NotImplementedError):
    @staticmethod
    def create(cls, attr):
        msg = "%s.%s is not yet implemented. You can help." % (cls, attr)
        return MongitaNotImplementedError(msg)

    @staticmethod
    def create_client(cls, attr):
        msg = "%s.%s is not yet implemented. Most MongoClient attributes/methods " \
              "will never be implemented because this is the key place where Mongita differs. " \
              "See the Mongita docs." % (cls, attr)
        return MongitaNotImplementedError(msg)

    @staticmethod
    def create_depr(cls, attr):
        msg = "%s.%s is deprecated and will not be implemented in Mongita." % (cls, attr)
        return MongitaNotImplementedError(msg)


class InvalidName(MongitaError):
    pass


class InvalidOperation(MongitaError):
    pass


class OperationFailure(MongitaError):
    pass


class DuplicateKeyError(MongitaError):
    pass


# For pymongo compatibility - especially in unit tests
PyMongoError = MongitaError



import json
import shutil
import unittest
from pathlib import Path

from retk import const, version_manager, __version__


class RecoverTest(unittest.TestCase):
    def setUp(self) -> None:
        self.tmp_dir = Path(__file__).parent / "tmp"
        self.tmp_dir.mkdir(parents=True, exist_ok=True)

    def tearDown(self) -> None:
        # remove all files and dirs
        shutil.rmtree(self.tmp_dir, ignore_errors=True)

    def test_package_version(self):
        vs = __version__.split(".")
        vs_int = list(map(int, vs))
        self.assertEqual(3, len(vs_int))

    def test_dump_load(self):
        version_manager.recover.dump_default_dot_rethink(self.tmp_dir / ".rethink.json")

        v = version_manager.recover.load_dot_rethink(self.tmp_dir / ".rethink.json")
        self.assertIsNotNone(v)
        # self.assertEqual(__version__, v["version"])
        self.assertEqual(const.DEFAULT_USER["email"], v["email"])
        self.assertEqual(const.DEFAULT_USER["nickname"], v["nickname"])
        self.assertEqual(const.DEFAULT_USER["avatar"], v["avatar"])
        self.assertEqual(const.DEFAULT_USER["email"], v["account"])
        self.assertEqual(const.LanguageEnum.EN.value, v["settings"]["language"])
        self.assertEqual(const.app.AppThemeEnum.LIGHT.value, v["settings"]["theme"])
        self.assertEqual(const.app.EditorModeEnum.WYSIWYG.value, v["settings"]["editorMode"])
        self.assertEqual(15, v["settings"]["editorFontSize"])
        self.assertEqual(const.app.EditorCodeThemeEnum.GITHUB.value, v["settings"]["editorCodeTheme"])
        self.assertEqual(200, v["settings"]["editorSepRightWidth"])
        self.assertEqual("", v["settings"]["editorSideCurrentToolId"])


class MigrateTest(unittest.TestCase):
    def setUp(self) -> None:
        self.tmp_dir = Path(__file__).parent / "tmp"
        self.tmp_dir.mkdir(parents=True, exist_ok=True)

    def tearDown(self) -> None:
        # remove all files and dirs
        shutil.rmtree(self.tmp_dir, ignore_errors=True)

    def test_merge_to_0_2_7(self):
        dot_rethink = {
            "_id": "65658d61dc7f58455d9b38b6",
            "id": "a" * 22,
            "email": "test@rethink.run",
            "nickname": "rethink",
            "avatar": "",
            "account": "test@rethink.run",
            "language": "en"
        }

        dot_data_dir = self.tmp_dir / const.settings.DOT_DATA
        dot_data_dir.mkdir(parents=True, exist_ok=True)

        with open(self.tmp_dir / const.settings.DOT_DATA / ".rethink.json", "w", encoding="utf-8") as f:
            json.dump(dot_rethink, f, indent=2, ensure_ascii=False)

        version_manager.migrate.to_latest_version(self.tmp_dir)
        v = version_manager.recover.load_dot_rethink(self.tmp_dir / const.settings.DOT_DATA / ".rethink.json")
        # self.assertEqual(__version__, v["version"])
        self.assertGreater(len(v["version"]), 0)
        self.assertIn("settings", v)



from typing import Any

from retk import config
from retk.depend.mongita.results import UpdateResult
from .client import client


async def remove_from_node(from_nid: str, to_nid: str):
    if config.is_local_db():
        # no $pull support
        to_n = await client.coll.nodes.find_one({"id": to_nid})
        if to_n is None:
            return
        try:
            to_n["fromNodeIds"].remove(from_nid)
            await client.coll.nodes.update_one(
                {"id": to_nid},
                {"$set": {"fromNodeIds": to_n["fromNodeIds"]}}
            )
        except ValueError:
            pass
    else:
        await client.coll.nodes.update_one(
            {"id": to_nid},
            {"$pull": {"fromNodeIds": from_nid}}
        )


async def node_add_to_set(id_: str, key: str, value: Any) -> UpdateResult:
    res = UpdateResult(0, 0)
    if config.is_local_db():
        # no $addToSet support
        has_new = False
        doc = await client.coll.nodes.find_one({"id": id_})
        if doc is None:
            return res
        if key not in doc:
            doc[key] = []
        if value not in doc[key]:
            doc[key].append(value)
            has_new = True
        if has_new:
            res = await client.coll.nodes.update_one(
                {"id": id_},
                {"$set": {key: doc[key]}}
            )
    else:
        res = await client.coll.nodes.update_one(
            {"id": id_},
            {"$addToSet": {key: value}}
        )
    return res


def sort_nodes_by_to_nids(condition: dict, page: int, limit: int):
    if config.is_local_db():
        docs = client.coll.nodes.find(condition).sort(
            [("toNodeIdsLen", -1), ("_id", -1)]
        ).skip(page * limit).limit(limit)
    else:
        docs = client.coll.nodes.aggregate([
            {"$match": condition},
            {"$addFields": {"toNodeIdsLen": {"$size": "$toNodeIds"}}},
            {"$sort": {"toNodeIdsLen": -1}},
            {"$sort": {"_id": -1}},
            {"$skip": page * limit},
            {"$limit": limit},
        ])
    return docs



from . import version_manager  # noqa: F401
from ._version import __version__  # noqa: F401
from .core import scheduler
from .models import tps  # noqa: F401
from .plugins.base import (  # noqa: F401
    Plugin,
    add_plugin,
    remove_plugin,
    PluginAPICallReturn,
)
from .run import run  # noqa: F401



import abc
import os
import pathlib

import bson

from .command_cursor import CommandCursor
from .common import support_alert, ok_name, MetaStorageObject
from .database import Database
from .engines import disk_engine, memory_engine
from .errors import MongitaNotImplementedError, InvalidName
from .read_concern import ReadConcern
from .write_concern import WriteConcern

DEFAULT_STORAGE_DIR = os.path.join(pathlib.Path.home(), '.mongita')


class MongitaClient(abc.ABC):
    UNIMPLEMENTED = ['HOST', 'PORT', 'address', 'arbiters', 'close', 'close_cursor',
                     'codec_options', 'database_names', 'event_listeners', 'fsync',
                     'get_database', 'get_default_database', 'is_locked',
                     'is_mongos', 'is_primary', 'kill_cursors',
                     'local_threshold_ms', 'max_bson_size', 'max_idle_time_ms',
                     'max_message_size', 'max_pool_size', 'max_write_batch_size',
                     'min_pool_size', 'next', 'nodes', 'primary',
                     'read_preference', 'retry_reads', 'retry_writes',
                     'secondaries', 'server_info', 'server_selection_timeout',
                     'set_cursor_manager', 'start_session', 'unlock', 'watch']
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self._base_location = ''
        self._cache = {}
        self._read_concern = ReadConcern()
        self._write_concern = WriteConcern()

    def __getattr__(self, attr):
        if attr in self.UNIMPLEMENTED:
            raise MongitaNotImplementedError.create_client("MongitaClient", attr)
        if attr == '_Database__create':
            return self.__create
        return self[attr]

    def __getitem__(self, db_name):
        try:
            return self._cache[db_name]
        except KeyError:
            if not ok_name(db_name):
                raise InvalidName("Database cannot be named %r." % db_name)
            db = Database(db_name, self)
            self._cache[db_name] = db
            return db

    def __create(self, db_name):
        """
        Mongodb does not create anything until first insert. So when we insert
        something, this will create a small metadata file to basically just store
        our database_names
        """
        metadata = self.engine.get_metadata(self._base_location)
        if metadata:
            if db_name not in metadata['database_names']:
                metadata['database_names'].append(db_name)
                assert self.engine.put_metadata(self._base_location, metadata)
            return
        metadata = MetaStorageObject({
            'options': {},
            'database_names': [db_name],
            'uuid': str(bson.ObjectId()),
        })
        assert self.engine.put_metadata(self._base_location, metadata)

    @property
    def read_concern(self):
        return self._read_concern

    @property
    def write_concern(self):
        return self._write_concern

    @support_alert
    async def close(self):
        """
        Close the connection to the database and remove all cache.

        :rtype: None
        """
        self.engine.close()

    @support_alert
    async def list_database_names(self):
        """
        List every database name.

        :rtype: list[str]
        """
        metadata = self.engine.get_metadata(self._base_location)
        if metadata:
            return metadata['database_names']
        return []

    @support_alert
    async def list_databases(self):
        """
        Returns a cursor to iterate over all databases.

        :rtype: CommandCursor
        """

        async def cursor():
            for db_name in await self.list_database_names():
                if db_name not in self._cache:
                    self._cache[db_name] = Database(db_name, self)
                yield self._cache[db_name]

        return CommandCursor(cursor)

    @support_alert
    async def drop_database(self, name_or_database):
        """
        Drop a database.

        :param name_or_database str|database.Database:
        :rtype: None
        """
        if isinstance(name_or_database, Database):
            db_name = name_or_database.name
        else:
            db_name = name_or_database

        db = self[db_name]
        for coll in list(db.list_collection_names()):
            db.drop_collection(coll)
        metadata = self.engine.get_metadata(self._base_location)
        if metadata and db_name in metadata['database_names']:
            metadata['database_names'].remove(db_name)
            assert self.engine.put_metadata(self._base_location, metadata)
        self.engine.delete_dir(db_name)
        self._cache.pop(db_name, None)


class MongitaClientDisk(MongitaClient):
    """
    The MongoClientDisk persists its state on the disk. It is meant to be
    compatible in most ways with pymongo's MongoClient.
    """

    def __init__(self, host=DEFAULT_STORAGE_DIR, **kwargs):
        host = host or DEFAULT_STORAGE_DIR
        if isinstance(host, list):
            # fix for mongoengine passing a list to us
            host = host[0]
        self.engine = disk_engine.DiskEngine.create(host)
        self.is_primary = True
        super().__init__()

    def __repr__(self):
        path = self.engine.base_storage_path
        return "MongitaClientDisk(path=%s)" % path


class MongitaClientMemory(MongitaClient):
    """
    The MongoClientMemory only holds its state in memory. Nonetheless, it is
    still thread-safe. It is meant to be compatible in most ways with
    pymongo's MongoClient.
    """

    def __init__(self, host=':memory:', strict=False, **kwargs):
        self.engine = memory_engine.MemoryEngine.create(strict)
        self.is_primary = True
        super().__init__()

    def __repr__(self):
        return "MongitaClientMemory()"



import os

from fastapi import APIRouter, HTTPException, FastAPI
from fastapi.responses import HTMLResponse, FileResponse, RedirectResponse
from fastapi.staticfiles import StaticFiles

from retk import const, config
from retk.logger import logger
from retk.routes import utils

r_prefix = "/r"
r_router = APIRouter(
    prefix=r_prefix,
    tags=["self_hosted"],
    responses={404: {"description": "Not found"}},
)

home_router = APIRouter(
    tags=["self_hosted_home"],
    responses={404: {"description": "Not found"}},
)

node_file_router = APIRouter(
    prefix="/files",
    tags=["self_hosted_files"],
    responses={404: {"description": "Not found"}},
)


@r_router.get("/", response_class=HTMLResponse)
@r_router.get("/sauth", response_class=HTMLResponse)
@r_router.get("/{path}", response_class=HTMLResponse)
@r_router.get("/plugin/{pid}", response_class=HTMLResponse)
@r_router.get("/n/{nid}", response_class=HTMLResponse)
async def app_page() -> HTMLResponse:
    if not config.is_local_db():
        raise HTTPException(status_code=404, detail="only support local storage")
    content = (const.settings.FRONTEND_DIR / "index.html").read_text(encoding="utf-8")
    api_url = os.getenv('VUE_APP_API_URL')
    script = f"window.VUE_APP_API_URL='{api_url}';"
    language = os.getenv("RETHINK_DEFAULT_LANGUAGE")
    if language is not None:
        script += f"window.VUE_APP_LANGUAGE='{language}';"
    if os.getenv("RETHINK_SERVER_PASSWORD", None) is not None:
        script += "window.VUE_APP_ONE_USER_REQUIRE_AUTH=1;"
    script = f"<script>{script}</script>"
    return HTMLResponse(
        content=content + script,
        status_code=200,
    )


@home_router.get("/", response_class=RedirectResponse)
async def index() -> RedirectResponse:
    return RedirectResponse(url=r_prefix)


# no /r/ needed, only use in local
@node_file_router.get(
    "/{fid}",
    status_code=200,
    response_class=FileResponse,
)
async def user_data(
        fid: str = utils.ANNOTATED_FID
) -> FileResponse:
    if not config.is_local_db():
        raise HTTPException(status_code=404, detail="only support local storage")
    prefix = config.get_settings().RETHINK_LOCAL_STORAGE_PATH
    return FileResponse(
        path=prefix / const.settings.DOT_DATA / "files" / fid,
        status_code=200,
    )


def mount_static(app: FastAPI):
    try:
        for name in ["css", "js", "img", "dist"]:
            app.mount(
                f"{r_prefix}/{name}",
                StaticFiles(directory=const.settings.FRONTEND_DIR / name),
                name=name,
            )
    except RuntimeError:
        logger.debug("mount frontend failed, the frontend files is in somewhere else")



from enum import IntEnum, unique


@unique
class UserBehaviorTypeEnum(IntEnum):
    LOGIN = 0  # backend
    LOGOUT = 1  # backend
    # node ops, backend
    NODE_CREATE = 2
    NODE_QUICK_CREATE = 3
    NODE_TRASHED_OPS = 4
    NODE_RESTORED_OPS = 5
    NODE_DELETED_OPS = 6
    # search, backend
    SEARCH_GLOBAL = 7
    SEARCH_AT = 8
    # at ops, frontend
    AT_NODE = 9
    AT_RECOMMENDED_NODE = 10
    # view node, frontend
    VIEW_NODE_FROM_RECOMMENDATION = 11
    VIEW_NODE_FROM_LINKED_NODE = 12


USER_BEHAVIOR_TYPE_MAP = {
    t.value: t
    for t in UserBehaviorTypeEnum.__members__.values()
}



from .common import ASCENDING, DESCENDING, support_alert
from .errors import MongitaNotImplementedError, MongitaError, InvalidOperation


def _validate_sort(key_or_list, direction=None):
    """
    Validate kwargs and return a proper sort list

    :param key_or_list str|[(str key, int direction), ...]
    :param direction int:
    :rtype: [(str key, int direction), ...]
    """
    if direction is None and isinstance(key_or_list, (list, tuple)) \
            and all(isinstance(tup, (list, tuple)) and len(tup) == 2 for tup in key_or_list):
        _sort = key_or_list
    elif direction is None and isinstance(key_or_list, str):
        _sort = [(key_or_list, ASCENDING)]
    elif isinstance(key_or_list, str) and isinstance(direction, int):
        _sort = [(key_or_list, direction)]
    else:
        raise MongitaError("Unsupported sort parameter format. See the docs.")
    for sort_key, sort_direction in _sort:
        if not isinstance(sort_key, str):
            raise MongitaError("Sort key(s) must be strings %r" % str(key_or_list))
        if sort_direction not in (ASCENDING, DESCENDING):
            raise MongitaError("Sort direction(s) must be either ASCENDING (1) or DESCENDING (-1). Not %r" % direction)
    return _sort


class Cursor():
    UNIMPLEMENTED = ['add_option', 'address', 'alive', 'allow_disk_use', 'batch_size',
                     'collation', 'collection', 'comment', 'cursor_id', 'distinct',
                     'explain', 'hint', 'max', 'max_await_time_ms',
                     'max_time_ms', 'min', 'remove_option', 'retrieved', 'rewind',
                     'session', 'where']
    DEPRECATED = ['count', 'max_scan']

    def __init__(self, _find, filter, sort, limit, skip):
        self._find = _find
        self._filter = filter
        self._sort = sort or []
        self._limit = limit or None
        self._skip = skip or None
        self._cursor = None

    def __getattr__(self, attr):
        if attr in self.DEPRECATED:
            raise MongitaNotImplementedError.create_depr("Collection", attr)
        if attr in self.UNIMPLEMENTED:
            raise MongitaNotImplementedError.create("Cursor", attr)
        raise AttributeError()

    def __getitem__(self, val):
        raise MongitaNotImplementedError.create("Cursor", '__getitem__')

    def __iter__(self):
        for el in self._gen():
            yield el

    def __next__(self):
        return next(self._gen())

    def _gen(self):
        """
        This exists so that we can maintain our position in the cursor and
        to not execute until we start requesting items
        """
        if self._cursor:
            return self._cursor
        self._cursor = self._find(filter=self._filter, sort=self._sort,
                                  limit=self._limit, skip=self._skip)
        return self._cursor

    @support_alert
    async def next(self):
        """
        Returns the next document in the Cursor. Raises StopIteration if there
        are no more documents.

        :rtype: dict
        """
        return next(self._gen())

    @support_alert
    def sort(self, key_or_list, direction=None):
        """
        Apply a sort to the cursor. Sorts have no impact until retrieving the
        first document from the cursor. If not sorting against indexes, sort can
        negatively impact performance.
        This returns the same cursor to allow for chaining. Only the last sort
        is applied.

        :param key_or_list str|[(key, direction)]:
        :param direction mongita.ASCENDING|mongita.DESCENDING:
        :rtype: cursor.Cursor
        """

        self._sort = _validate_sort(key_or_list, direction)
        if self._cursor:
            raise InvalidOperation("Cursor has already started and can't be sorted")

        return self

    @support_alert
    def limit(self, limit):
        """
        Apply a limit to the number of elements returned from the cursor.
        This returns the same cursor to allow for chaining. Only the last limit
        is applied.

        :param limit int:
        :rtype: cursor.Cursor
        """
        if not isinstance(limit, int):
            raise TypeError('Limit must be an integer')

        if self._cursor:
            raise InvalidOperation("Cursor has already started and can't be limited")

        self._limit = limit
        return self

    @support_alert
    def skip(self, skip):
        """
        Skip the first [skip] results of this cursor.
        """
        if not isinstance(skip, int):
            raise TypeError("The 'skip' parameter must be an integer")
        if skip < 0:
            raise ValueError("The 'skip' parameter must be >=0")
        if self._cursor:
            raise InvalidOperation("Cursor has already started and skip can't be applied")

        self._skip = skip
        return self

    @support_alert
    def clone(self):
        return Cursor(self._find, self._filter, self._sort, self._limit, self._skip)

    @support_alert
    def close(self):
        """
        Close this cursor to free the memory
        """
        self._cursor = iter(())

    async def to_list(self, length: int) -> list:
        if length is None:
            res = []
            while True:
                try:
                    res.append(next(self._gen()))
                except StopIteration:
                    break
        else:
            res = [next(self._gen()) for _ in range(0, length)]
        return res
